#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype code-based object data validation unit tests.**

This submodule unit tests the subset of the public API of the
:mod:`beartype.vale` subpackage defined by the private
:mod:`beartype.vale._is._valeisobj` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ CONSTANTS                         }....................
AS_WHEN_NIGHT_IS_BARE = (
    "From one lonely cloud",
    "The moon rains out her beams, and Heaven is overflow'd.",
)
'''
Arbitrary tuple to be assigned to an arbitrary instance variable of the
arbitrary class defined below.
'''

# ....................{ CLASSES ~ good                    }....................
class AllTheEarthAndAir(object):
    '''
    Arbitrary class defining an arbitrary attribute whose value has *no*
    accessible attributes but satisfies a validator tested below.
    '''

    def __init__(self) -> None:
        '''
        Initialize this object by defining this attribute.
        '''

        # Initialize this attribute to a shallow copy of this list rather
        # than this list itself to properly test equality comparison.
        self.with_thy_voice_is_loud = AS_WHEN_NIGHT_IS_BARE[:]


class WhatThouArtWeKnowNot(object):
    '''
    Arbitrary class defining an arbitrary attribute whose value is an instance
    of another class defining another arbitrary attribute.
    '''

    def __init__(self) -> None:
        '''
        Initialize this object by defining this attribute.
        '''

        self.what_is_most_like_thee = AllTheEarthAndAir()

# ....................{ CLASSES ~ bad                     }....................
class InAPalaceTower(object):
    '''
    Arbitrary class defining an arbitrary attribute of a differing name as that
    defined by the :class:`AllTheEarthAndAir` class.
    '''

    def __init__(self) -> None:
        '''
        Initialize this object by defining this attribute.
        '''

        self.soothing_her_love_laden = [
            'Soul in secret hour',
            'With music sweet as love, which overflows her bower:',
        ]


class TillTheWorldIsWrought(object):
    '''
    Arbitrary class defining an arbitrary attribute of the same name as that
    defined by the :class:`AllTheEarthAndAir` class whose value has *no*
    accessible attributes and does *not* satisfy a validator tested below.
    '''

    def __init__(self) -> None:
        '''
        Initialize this object by defining this attribute.
        '''

        # Initialize this attribute to an arbitrary list unequal to the value
        # initializing the comparable
        # "AllTheEarthAndAir.with_thy_voice_is_loud" attribute.
        self.with_thy_voice_is_loud = [
            'To sympathy with hopes and fears it heeded not:',
            'As from thy presence showers a rain of melody.',
        ]


class InADellOfDew(object):
    '''
    Arbitrary class defining an arbitrary attribute of the same name as that
    defined by the :class:`WhatThouArtWeKnowNot` class whose value is an
    instance of another class defining an attribute of the same name as that
    defined by the :class:`AllTheEarthAndAir` class whose value does *not*
    satisfy a validator tested below.
    '''

    def __init__(self) -> None:
        '''
        Initialize this object by defining this attribute.
        '''

        self.what_is_most_like_thee = TillTheWorldIsWrought()

# ....................{ TESTS ~ class : isattr            }....................
def test_api_vale_isattr_pass() -> None:
    '''
    Test successful usage of the :mod:`beartype.vale.IsAttr` factory.
    '''

    # Defer test-specific imports.
    from beartype.vale import IsAttr, IsEqual
    from beartype.vale._core._valecore import BeartypeValidator

    # Instances of valid test classes declared above.
    from_rainbow_clouds = AllTheEarthAndAir()
    drops_so_bright_to_see = WhatThouArtWeKnowNot()

    # Instances of invalid test classes declared above.
    like_a_glow_worm_golden = InAPalaceTower()
    like_a_high_born_maiden = TillTheWorldIsWrought()
    scattering_unbeholden = InADellOfDew()

    # Validator produced by subscripting this factory with the name of an
    # attribute defined by the former class and that attribute's value.
    IsInTheLightOfThought = IsAttr[
        'with_thy_voice_is_loud', IsEqual[AS_WHEN_NIGHT_IS_BARE]]

    # Validator produced by subscripting this factory with the name of an
    # attribute defined by the latter class and the prior validator.
    IsSingingHymnsUnbidden = IsAttr[
        'what_is_most_like_thee', IsInTheLightOfThought]

    # Assert these validators satisfy the expected API.
    assert isinstance(IsInTheLightOfThought, BeartypeValidator)
    assert isinstance(IsSingingHymnsUnbidden, BeartypeValidator)

    # Assert these validators produce the same objects when subscripted by the
    # same arguments (and are thus memoized on subscripted arguments).
    assert IsSingingHymnsUnbidden is IsAttr[
        'what_is_most_like_thee', IsInTheLightOfThought]

    # Assert these validators accept objects defining attributes with the same
    # names and values as those subscripting these validators.
    assert IsInTheLightOfThought.is_valid(from_rainbow_clouds) is True
    assert IsSingingHymnsUnbidden.is_valid(drops_so_bright_to_see) is True

    # Assert these validators reject objects defining attributes with differing
    # names to those subscripting these validators.
    assert IsInTheLightOfThought.is_valid(like_a_glow_worm_golden) is False
    assert IsSingingHymnsUnbidden.is_valid(like_a_glow_worm_golden) is False

    # Assert these validators reject objects defining attributes with the same
    # names but differing values to those subscripting these validators.
    assert IsInTheLightOfThought.is_valid(like_a_high_born_maiden) is False
    assert IsSingingHymnsUnbidden.is_valid(scattering_unbeholden) is False

    # Assert these validators have the expected representation.
    IsInTheLightOfThought_repr = repr(IsInTheLightOfThought)
    assert repr('with_thy_voice_is_loud') in IsInTheLightOfThought_repr
    assert repr(AS_WHEN_NIGHT_IS_BARE) in IsInTheLightOfThought_repr

    # Validator synthesized from the above validators with the domain-specific
    # language (DSL) supported by these validators.
    IsInTheLightOfThoughtOrSingingHymnsUnbidden = (
        IsInTheLightOfThought | IsSingingHymnsUnbidden)

    # Assert this object performs the expected validation.
    assert IsInTheLightOfThoughtOrSingingHymnsUnbidden.is_valid(
        AllTheEarthAndAir()) is True
    assert IsInTheLightOfThoughtOrSingingHymnsUnbidden.is_valid(
        WhatThouArtWeKnowNot()) is True
    assert IsInTheLightOfThoughtOrSingingHymnsUnbidden.is_valid(
        InAPalaceTower()) is False
    assert IsInTheLightOfThoughtOrSingingHymnsUnbidden.is_valid(
        TillTheWorldIsWrought()) is False
    assert IsInTheLightOfThoughtOrSingingHymnsUnbidden.is_valid(
        InADellOfDew()) is False

    # Assert this object provides the expected representation.
    assert '|' in repr(IsInTheLightOfThoughtOrSingingHymnsUnbidden)


def test_api_vale_isattr_fail() -> None:
    '''
    Test unsuccessful usage of the :mod:`beartype.vale.IsAttr` factory.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeValeSubscriptionException
    from beartype.vale import IsAttr, IsEqual
    from pytest import raises

    # Assert that subscripting this factory with the empty tuple raises the
    # expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr[()]

    # Assert that subscripting this factory with the empty tuple raises the
    # expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr[()]

    # Assert that subscripting this factory with one non-tuple argument raises
    # the expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr['Among the flowers and grass, which screen it from the view:']

    # Assert that subscripting this factory with three or more arguments raises
    # the expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr[
            "Like a rose embower'd",
            "In its own green leaves,",
            "By warm winds deflower'd,",
        ]

    # Assert that subscripting this factory with two arguments whose first
    # argument is *NOT* a string raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr[
            IsEqual['Till the scent it gives'],
            IsEqual[
                'Makes faint with too much sweet those heavy-winged thieves:']]

    # Assert that subscripting this factory with two arguments whose first
    # argument is the empty string raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr['', IsEqual['Sound of vernal showers']]

    # Assert that subscripting this factory with two arguments whose first
    # argument is an invalid Python identifier raises the expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsAttr['On the twinkling grass,', IsEqual["Rain-awaken'd flowers,"]]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python module tester** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.module.utilmodtest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ raiser                     }....................
def test_die_unless_module_attr_name() -> None:
    '''
    Test the :func:`beartype._util.module.utilmodtest.die_unless_module_attr_name`
    validator.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilModuleException
    from beartype._util.module.utilmodtest import die_unless_module_attr_name
    from pytest import raises

    # Assert this validator raises *NO* exception when passed a syntactically
    # valid Python identifier.
    die_unless_module_attr_name('She_dwelt.among_the.untrodden.ways_2')

    # Assert this validator raises the expected exception when passed:
    # * A non-string.
    # * The empty string.
    # * A non-empty string containing *NO* "." delimiters.
    # * A non-empty string syntactically invalid as a Python identifier.
    with raises(_BeartypeUtilModuleException):
        die_unless_module_attr_name(b'Sentient.No6')
    with raises(_BeartypeUtilModuleException):
        die_unless_module_attr_name('')
    with raises(_BeartypeUtilModuleException):
        die_unless_module_attr_name('typing')
    with raises(_BeartypeUtilModuleException):
        die_unless_module_attr_name('Sentient.6')

# ....................{ TESTS ~ tester                     }....................
def test_is_module() -> None:
    '''
    Test the :func:`beartype._util.module.utilmodtest.is_module` tester.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeModuleUnimportableWarning
    from beartype._util.module.utilmodtest import is_module
    from pytest import warns

    # Assert this tester accepts the name of a (possibly unimported) existing
    # importable module.
    assert is_module(
        'beartype_test.a00_unit.data.util.mod.data_utilmodule_good') is True

    # Assert this tester accepts the name of an already imported module.
    assert is_module(
        'beartype_test.a00_unit.data.util.mod.data_utilmodule_good') is True

    # Assert this tester rejects the name of a module guaranteed *NOT* to
    # exist, because we fully control the "beartype_test" package.
    assert is_module(
        'beartype_test.a00_unit.data.util.mod.data_utilmodule_nonexistent'
    ) is False

    # Assert this function emits the expected warning when passed the name of
    # an existing unimportable module.
    with warns(BeartypeModuleUnimportableWarning):
        assert is_module(
            'beartype_test.a00_unit.data.util.mod.data_utilmodule_bad') is (
            False)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python version-agnostic signs** (i.e., instances of the
:class:`beartype._data.hint.pep.sign.datapepsigncls.HintSign` class
uniquely identifying PEP-compliant type hints in a safe, non-deprecated manner
regardless of the Python version targeted by the active Python interpreter).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Embed the number of each PEP for PEP-specific signs into the names of
#those signs (e.g., rename "HintSignAnnotated" to "HintSignPep593Annotated").

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: Attributes imported here at module scope *MUST* be explicitly
# deleted from this module's namespace below.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype._data.hint.pep.sign.datapepsigncls import HintSign as _HintSign

# ....................{ SIGNS ~ explicit                   }....................
# Signs with explicit analogues in the stdlib "typing" module.
#
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: Signs defined by this module are synchronized with the "__all__"
# list global of the "typing" module bundled with the most recent CPython
# release. For that reason, these signs are:
# * Intentionally declared in the exact same order prefixed by the exact same
#   inline comments as for that list global.
# * Intentionally *NOT* commented with docstrings, both because:
#   * These docstrings would all trivially reduce to a single-line sentence
#     fragment resembling "Alias of typing attribute."
#   * These docstrings would inhibit diffing and synchronization by inspection.
# * Intentionally *NOT* conditionally isolated to the specific range of Python
#   versions whose "typing" module lists these attributes. For example, the
#   "HintSignAsyncContextManager" sign identifying the
#   "typing.AsyncContextManager" attribute that only exists under Python >=
#   3.7 could be conditionally isolated to that range of Python versions.
#   Technically, there exists *NO* impediment to doing so; pragmatically, doing
#   so would be ineffectual. Why? Because attributes *NOT* defined by the
#   "typing" module of the active Python interpreter cannot (by definition) be
#   used to annotate callables decorated by the @beartype decorator.
#
# When bumping beartype to support a new CPython release:
# * Declare one new attribute here for each new "typing" attribute added by
#   that CPython release regardless of whether beartype explicitly supports
#   that attribute yet. The subsequently called die_unless_hint_pep_supported()
#   validator will raise exceptions when passed these attributes.
# * Preserve attributes here that have since been removed from the "typing"
#   module in that CPython release to ensure their continued usability when
#   running beartype against older CPython releases.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# Super-special typing primitives.
HintSignAnnotated = _HintSign(name='Annotated')
HintSignAny = _HintSign(name='Any')
HintSignCallable = _HintSign(name='Callable')
HintSignClassVar = _HintSign(name='ClassVar')
HintSignConcatenate = _HintSign(name='Concatenate')
HintSignFinal = _HintSign(name='Final')
HintSignForwardRef = _HintSign(name='ForwardRef')
HintSignGeneric = _HintSign(name='Generic')
HintSignLiteral = _HintSign(name='Literal')
HintSignOptional = _HintSign(name='Optional')
HintSignParamSpec = _HintSign(name='ParamSpec')
HintSignProtocol = _HintSign(name='Protocol')
HintSignTuple = _HintSign(name='Tuple')
HintSignType = _HintSign(name='Type')
HintSignTypeVar = _HintSign(name='TypeVar')
HintSignTypeVarTuple = _HintSign(name='TypeVarTuple')
HintSignUnion = _HintSign(name='Union')

# ABCs (from collections.abc).
HintSignAbstractSet = _HintSign(name='AbstractSet')
HintSignByteString = _HintSign(name='ByteString')
HintSignContainer = _HintSign(name='Container')
HintSignContextManager = _HintSign(name='ContextManager')
HintSignHashable = _HintSign(name='Hashable')
HintSignItemsView = _HintSign(name='ItemsView')
HintSignIterable = _HintSign(name='Iterable')
HintSignIterator = _HintSign(name='Iterator')
HintSignKeysView = _HintSign(name='KeysView')
HintSignMapping = _HintSign(name='Mapping')
HintSignMappingView = _HintSign(name='MappingView')
HintSignMutableMapping = _HintSign(name='MutableMapping')
HintSignMutableSequence = _HintSign(name='MutableSequence')
HintSignMutableSet = _HintSign(name='MutableSet')
HintSignSequence = _HintSign(name='Sequence')
HintSignSized = _HintSign(name='Sized')
HintSignValuesView = _HintSign(name='ValuesView')
HintSignAwaitable = _HintSign(name='Awaitable')
HintSignAsyncIterator = _HintSign(name='Iterator')
HintSignAsyncIterable = _HintSign(name='Iterable')
HintSignCoroutine = _HintSign(name='Coroutine')
HintSignCollection = _HintSign(name='Collection')
HintSignAsyncGenerator = _HintSign(name='AsyncGenerator')
HintSignAsyncContextManager = _HintSign(name='ContextManager')

# Structural checks, a.k.a. protocols.
HintSignReversible = _HintSign(name='Reversible')
# SupportsAbs   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsBytes   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsComplex   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsFloat   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsIndex   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsInt   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsRound   <-- not a useful type hint (already an isinstanceable ABC)

# Concrete collection types.
HintSignChainMap = _HintSign(name='ChainMap')
HintSignCounter = _HintSign(name='Counter')
HintSignDeque = _HintSign(name='Deque')
HintSignDict = _HintSign(name='Dict')
HintSignDefaultDict = _HintSign(name='DefaultDict')
HintSignList = _HintSign(name='List')
HintSignOrderedDict = _HintSign(name='OrderedDict')
HintSignSet = _HintSign(name='Set')
HintSignFrozenSet = _HintSign(name='FrozenSet')
HintSignNamedTuple = _HintSign(name='NamedTuple')
HintSignTypedDict = _HintSign(name='TypedDict')
HintSignGenerator = _HintSign(name='Generator')

# Other concrete types.
HintSignMatch = _HintSign(name='Match')
HintSignPattern = _HintSign(name='Pattern')

# Other concrete type aliases.
HintSignIO = HintSignGeneric
HintSignBinaryIO = HintSignGeneric
HintSignTextIO = HintSignGeneric

# One-off things.
# AnyStr   <-- not a unique type hint (just a constrained "TypeVar")
# cast   <-- unusable as a type hint
# final   <-- unusable as a type hint
# get_args   <-- unusable as a type hint
# get_origin   <-- unusable as a type hint
# get_type_hints   <-- unusable as a type hint
# is_typeddict   <-- unusable as a type hint
HintSignLiteralString = _HintSign(name='LiteralString')
HintSignNever = _HintSign(name='Never')
HintSignNewType = _HintSign(name='NewType')
# no_type_check   <-- unusable as a type hint
# no_type_check_decorator   <-- unusable as a type hint

# Note that "NoReturn" is contextually valid *ONLY* as a top-level return hint.
# Since this use case is extremely limited, we explicitly generate code for this
# use case outside of the general-purpose code generation pathway for standard
# type hints. Since "NoReturn" is an unsubscriptable singleton, we explicitly
# detect this type hint with an identity test and thus require *NO* sign to
# uniquely identify this type hint.
#
# Theoretically, explicitly defining a sign uniquely identifying this type hint
# could erroneously encourage us to use that sign elsewhere; we should avoid
# that, as "NoReturn" is invalid in almost all possible contexts. Pragmatically,
# doing so nonetheless improves orthogonality when detecting and validating
# PEP-compliant type hints, which ultimately matters more than our subjective
# feelings about the matter. Wisely, we choose pragmatics.
#
# In short, "NoReturn" is insane.
HintSignNoReturn = _HintSign(name='NoReturn')

HintSignNotRequired = _HintSign(name='NotRequired')
# overload   <-- unusable as a type hint
HintSignParamSpecArgs = _HintSign(name='ParamSpecArgs')
HintSignParamSpecKwargs = _HintSign(name='ParamSpecKwargs')
HintSignRequired = _HintSign(name='Required')
# runtime_checkable   <-- unusable as a type hint
HintSignSelf = _HintSign(name='Self')
# Text   <-- not actually a type hint (literal alias for "str")
# TYPE_CHECKING   <-- unusable as a type hint
HintSignTypeAlias = _HintSign(name='TypeAlias')
HintSignTypeGuard = _HintSign(name='TypeGuard')
HintSignUnpack = _HintSign(name='Unpack')

# Wrapper namespace for re type aliases.
#
# Note that "typing.__all__" intentionally omits the "Match" and "Pattern"
# attributes, which it oddly considers to comprise another namespace. *shrug*

# ....................{ SIGNS ~ implicit                   }....................
# Signs with *NO* explicit analogues in the stdlib "typing" module but
# nonetheless standardized by one or more PEPs.

HintSignNone = _HintSign(name='None')
'''
:pep:`484` explicitly supports the :data:`None` singleton, albeit implicitly:

    When used in a type hint, the expression None is considered equivalent to
    type(None).
'''

# ....................{ SIGNS ~ implicit : lib             }....................
# Signs identifying PEP-noncompliant third-party type hints published by...
#
# ....................{ SIGNS ~ implicit : lib : numpy     }....................
HintSignNumpyArray = _HintSign(name='NumpyArray')   # <-- "numpy.typing.NDArray"
'''
...the :mod:`numpy.typing` subpackage.
'''

# ....................{ SIGNS ~ implicit : lib : pandera   }....................
HintSignPanderaAny = _HintSign(name='PanderaAny')   # <-- "pandera.typing.*"
'''
...the :mod:`pandera.typing` subpackage.

Specifically, define a single sign unconditionally matching *all* type hints
published by the :mod:`pandera.typing` subpackage. Why? Because Pandera insanely
publishes its own Pandera-specific PEP-noncompliant runtime type-checking
decorator :func:`pandera.check_types` that supports *only* Pandera-specific
PEP-noncompliant :mod:`pandera.typing` type hints. Since Pandera users are
already accustomed to decorating *all* Pandera-based callables (i.e., callables
accepting one or more parameters and/or returning one or more values which are
Pandera objects) by :func:`pandera.check_types`, attempting to type-check the
same objects already type-checked by that decorator would only inefficiently and
needlessly slow :mod:`beartype` down. Ergo, we ignore *all* Pandera type hints
by:

* Defining this catch-all singleton for Pandera type hints here.
* Denoting this singleton to be unconditionally ignorable elsewhere.
'''

# ....................{ SIGNS ~ implicit : pep : 557       }....................
# dataclasses.InitVar[...].
HintSignPep557DataclassInitVar = _HintSign(name='Pep557DataclassInitVar')
'''
:pep:`557`-compliant :obj:`dataclasses.InitVar` type hint factory, annotating
class-scoped variable annotations of :func:`dataclass.dataclass`-decorated
data classes.
'''

# ....................{ SIGNS ~ implicit : pep : 585       }....................
# os.PathLike[...], weakref.weakref[...], et al.
HintSignPep585BuiltinSubscriptedUnknown = _HintSign(
    name='Pep585BuiltinSubscriptedUnknown')
'''
:pep:`585`-compliant C-based :class:`types.GenericAlias` superclass inheritable
by PEP-noncompliant pure-Python subclasses in either the standard library or
third-party packages, which when subscripted by otherwise PEP-compliant child
type hints produce PEP-noncompliant **unrecognized subscripted builtin type
hints** (i.e., C-based type hints that are *not* isinstanceable types,
instantiated by subscripting pure-Python origin classes unrecognized by
:mod:`beartype` and thus PEP-noncompliant).

Examples include:

* ``os.PathLike[...]`` type hints.
* ``weakref.weakref[...]`` type hints.

Unsurprisingly, :mod:`beartype` reduces C-based unrecognized subscripted builtin
type hints (which are *not* type-checkable as is) to their unsubscripted
pure-Python origin classes (which are type-checkable as is).
'''

# ....................{ SIGNS ~ implicit : pep : 695       }....................
# "type {alias_name} = {alias_value}" statements.
HintSignPep695TypeAlias = _HintSign(name='HintSignPep695TypeAlias')
'''
:pep:`695`-compliant C-based :class:`types.TypeAliasType` class of all
:pep:`695`-compliant **type aliases** (i.e., objects created as the left-hand
sides of statements of the form ``type {alias_name} = {alias_value}``).
'''

# ....................{ CLEANUP                            }....................
# Prevent all attributes imported above from polluting this namespace. Why?
# Logic elsewhere subsequently assumes a one-to-one mapping between the
# attributes of this namespace and signs.
del _HintSign



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`673`-compliant :obj:`typing.Self` **unit tests**.

This submodule unit tests :pep:`673` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.11.0')
def test_decor_pep673() -> None:
    '''
    Test :pep:`673` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.11 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.door import (
        die_if_unbearable,
        is_bearable,
    )
    from beartype.roar import (
        BeartypeCallHintReturnViolation,
        BeartypeDecorHintPep673Exception,
    )
    from beartype.typing import (
        List,
        Self,
    )
    from pytest import raises

    # ....................{ CLASSES                        }....................
    @beartype
    class GoodClassIsGood(object):
        '''
        Arbitrary class decorated by :func:`.beartype`.
        '''

        def wonderful_method_is_wonderful(self: Self, count: int) -> List[Self]:
            '''
            Arbitrary method annotated by one or more :pep:`673`-compliant self
            type hints, guaranteed to *not* violate runtime type-checking.
            '''

            return [self] * count


        def horrible_method_is_horrible(self: Self) -> Self:
            '''
            Arbitrary method annotated by one or more :pep:`673`-compliant self
            type hints, guaranteed to violate runtime type-checking.
            '''

            return 'Cleave themselves into chasms, while far below'

    # Do not taunt Super Happy Fun Instance.
    super_happy_fun_instance = GoodClassIsGood()

    # ....................{ PASS                           }....................
    # Assert that a method of a @beartype-decorated class satisfying a PEP
    # 673-compliant self type hint returns the expected object.
    avoid_prolonged_exposure = (
        super_happy_fun_instance.wonderful_method_is_wonderful(count=42))
    assert len(avoid_prolonged_exposure) == 42
    assert avoid_prolonged_exposure[ 0] is super_happy_fun_instance
    assert avoid_prolonged_exposure[-1] is super_happy_fun_instance

    # Assert that @beartype raises the expected violation when calling a
    # method of a @beartype-decorated class erroneously violating a PEP
    # 673-compliant self type hint.
    with raises(BeartypeCallHintReturnViolation):
        super_happy_fun_instance.horrible_method_is_horrible()

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating
    # non-class objects accepting one or more parameters erroneously annotated
    # by a PEP 673-compliant self type hint. Self type hints are valid *ONLY*
    # inside classes decorated by @beartype.
    with raises(BeartypeDecorHintPep673Exception):
        @beartype
        def terribad_function_is_terribad(self: Self) -> Self:
            return self
    with raises(BeartypeDecorHintPep673Exception):
        class BadClassIsBad(object):
            @beartype
            def awful_method_is_awful(self: Self) -> Self:
                return self

    # Assert that statement-level runtime type-checkers raise the expected
    # exceptions when passed PEP 673-compliant self type hints.
    with raises(BeartypeDecorHintPep673Exception):
        die_if_unbearable(
            'So sweet, the sense faints picturing them! Thou', Self)
    with raises(BeartypeDecorHintPep673Exception):
        is_bearable(
            "For whose path the Atlantic's level powers", Self)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype plugin mixin unit tests.**

This submodule unit tests protocols defined by the private
:func:`beartype.plug._plughintable` submodule, most of which are publicly exported
to end users and thus critically important.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_beartypehintable() -> None:
    '''
    Test the :class:`beartype.plug._plughintable.BeartypeHintable` mixin.
    '''

    # .....................{ IMPORTS                       }....................
    # Defer test-specific imports.
    from beartype.plug import BeartypeHintable
    from pytest import raises

    # .....................{ CLASSES                       }....................
    class NorWhenTheFlakesBurn(BeartypeHintable):
        '''
        Arbitrary class explicitly satisfying the :class:`BeartypeHintable`
        protocol.
        '''

        @classmethod
        def __beartype_hint__(cls) -> object:
            '''
            Arbitrary beartype type hint transform reducing to an arbitrary
            PEP-compliant type hint.
            '''

            return str

    # .....................{ ASSERTS                       }....................
    # Assert that calling a concrete __beartype_hint__() class method of a
    # "BeartypeHintable" subclass returns the expected type hint.
    assert NorWhenTheFlakesBurn.__beartype_hint__() is str

    # Assert that calling the abstract __beartype_hint__() class method raises
    # the expected exception.
    with raises(NotImplementedError):
        BeartypeHintable.__beartype_hint__()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable tester utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.utilfunc.utilfunctest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
# Note that the related test_is_func_nested() unit test is intentionally defined
# in the companion "test_utilfuncscope" submodule rather than this submodule.
# Why? Convenience, mostly. That submodule defines various callables of interest
# that dramatically simplify the implementation of that unit test.

def test_is_func_lambda() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_lambda` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_lambda

    def intimations_of_immortality(): 'from Recollections of Early Childhood'

    # Assert this tester accepts pure-Python lambda functions.
    assert is_func_lambda(lambda: True) is True

    # Assert this tester rejects pure-Python non-lambda callables.
    assert is_func_lambda(intimations_of_immortality) is False

    # Assert this tester rejects C-based callables.
    assert is_func_lambda(iter) is False

# ....................{ TESTS ~ async                      }....................
def test_is_func_async() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_async` tester.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.utilfunctest import is_func_async
    from beartype_test.a00_unit.data.data_type import (
        async_generator,
        async_generator_factory,
        async_coroutine,
        async_coroutine_factory,
        sync_generator,
        sync_generator_factory,
        function,
    )

    # Assert this tester accepts pure-Python coroutine callables.
    assert is_func_async(async_coroutine_factory) is True

    # Assert this tester rejects pure-Python coroutine objects.
    assert is_func_async(async_coroutine) is False

    # Assert this tester accepts pure-Python asynchronous generator callables.
    assert is_func_async(async_generator_factory) is True

    # Assert this tester rejects pure-Python asynchronous generator objects.
    assert is_func_async(async_generator) is False

    # Assert this tester rejects pure-Python synchronous generator callables.
    assert is_func_async(sync_generator_factory) is False

    # Assert this tester rejects pure-Python synchronous generator objects.
    assert is_func_async(sync_generator) is False

    # Assert this tester rejects pure-Python non-asynchronous callables.
    assert is_func_async(function) is False

    # Assert this tester rejects arbitrary non-asynchronous objects.
    assert is_func_async('To hear—an old and solemn harmony;') is False


def test_is_func_coro() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_coro` function.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_coro
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype_test.a00_unit.data.data_type import (
        async_generator,
        async_generator_factory,
        async_coroutine,
        async_coroutine_factory,
        function,
    )

    # Assert this tester accepts pure-Python coroutine callables.
    assert is_func_coro(async_coroutine_factory) is True

    # Assert this tester rejects pure-Python coroutine objects.
    assert is_func_coro(async_coroutine) is False

    # Assert this tester rejects pure-Python asynchronous generator callables.
    assert is_func_coro(async_generator_factory) is False

    # Assert this tester rejects pure-Python asynchronous generator objects.
    assert is_func_coro(async_generator) is False

    # Assert this tester rejects pure-Python non-asynchronous callables.
    assert is_func_coro(function) is False

    # Assert this tester rejects arbitrary non-asynchronous objects.
    assert is_func_coro('To hear—an old and solemn harmony;') is (
        False)


def test_is_func_async_generator() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_async_generator` function.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_async_generator
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype_test.a00_unit.data.data_type import (
        async_coroutine,
        async_coroutine_factory,
        async_generator,
        async_generator_factory,
        sync_generator,
        sync_generator_factory,
        function,
    )

    # Assert this tester accepts pure-Python asynchronous generator callables.
    assert is_func_async_generator(async_generator_factory) is True

    # Assert this tester rejects pure-Python asynchronous generator objects.
    assert is_func_async_generator(async_generator) is False

    # Assert this tester rejects pure-Python coroutine callables.
    assert is_func_async_generator(async_coroutine_factory) is False

    # Assert this tester rejects pure-Python coroutine objects.
    assert is_func_async_generator(async_coroutine) is False

    # Assert this tester rejects pure-Python synchronous generator callables.
    assert is_func_async_generator(sync_generator_factory) is False

    # Assert this tester rejects pure-Python synchronous generator objects.
    assert is_func_async_generator(sync_generator) is False

    # Assert this tester rejects pure-Python non-asynchronous callables.
    assert is_func_async_generator(function) is False

    # Assert this tester rejects arbitrary non-asynchronous objects.
    assert is_func_async_generator('To hear—an old and solemn harmony;') is (
        False)

# ....................{ TESTS ~ sync                       }....................
def test_is_func_sync_generator() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_sync_generator` function.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_sync_generator
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype_test.a00_unit.data.data_type import (
        async_coroutine,
        async_coroutine_factory,
        async_generator,
        async_generator_factory,
        sync_generator,
        sync_generator_factory,
        function,
    )

    # Assert this tester rejects pure-Python asynchronous generator callables.
    assert is_func_sync_generator(async_generator_factory) is False

    # Assert this tester rejects pure-Python asynchronous generator objects.
    assert is_func_sync_generator(async_generator) is False

    # Assert this tester rejects pure-Python coroutine callables.
    assert is_func_sync_generator(async_coroutine_factory) is False

    # Assert this tester rejects pure-Python coroutine objects.
    assert is_func_sync_generator(async_coroutine) is False

    # Assert this tester accepts pure-Python synchronous generator callables.
    assert is_func_sync_generator(sync_generator_factory) is True

    # Assert this tester accepts pure-Python synchronous generator objects.
    assert is_func_sync_generator(sync_generator) is False

    # Assert this tester rejects pure-Python non-asynchronous callables.
    assert is_func_sync_generator(function) is False

    # Assert this tester rejects arbitrary non-asynchronous objects.
    assert is_func_sync_generator('To hear—an old and solemn harmony;') is (
        False)

# ....................{ TESTS ~ descriptor : classmethod   }....................
def test_die_unless_func_classmethod() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.die_unless_func_classmethod`
    validator.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.utilfunctest import die_unless_func_classmethod
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        Class,
    )
    from pytest import raises

    # Assert this validator accepts a class method descriptor.
    #
    # Note that class method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as class or instance
    # attributes, class methods reduce to instances of the standard
    # "beartype.cave.MethodBoundInstanceOrClassType" type.
    die_unless_func_classmethod(Class.__dict__['class_method'])

    # Assert this validator rejects *ALL* other callables.
    for some_callable in CALLABLES:
        with raises(_BeartypeUtilCallableException):
            die_unless_func_classmethod(some_callable)


def test_is_func_classmethod() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_classmethod` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_classmethod
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        Class,
    )

    # Assert this tester accepts a class method descriptor.
    #
    # Note that class method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as class or instance
    # attributes, class methods reduce to instances of the standard
    # "beartype.cave.MethodBoundInstanceOrClassType" type.
    assert is_func_classmethod(Class.__dict__['class_method']) is True

    # Assert this tester rejects *ALL* other callables.
    for some_callable in CALLABLES:
        assert is_func_classmethod(some_callable) is False

# ....................{ TESTS ~ descriptor : property      }....................
def test_die_unless_func_property() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.die_unless_func_property`
    validator.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.utilfunctest import die_unless_func_property
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        Class,
    )
    from pytest import raises

    # Assert this validator accepts a property descriptor.
    #
    # Note that property descriptors are directly accessible both as class
    # attributes *AND* via the low-level "object.__dict__" dictionary. Property
    # objects are *NOT* accessible as instance attributes, for obvious reasons.
    die_unless_func_property(Class.instance_property)

    # Assert this validator rejects *ALL* other callables.
    for some_callable in CALLABLES:
        with raises(_BeartypeUtilCallableException):
            die_unless_func_property(some_callable)


def test_is_func_property() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_property` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_property
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        Class,
    )

    # Assert this tester accepts a property descriptor.
    #
    # Note that property descriptors are directly accessible both as class
    # attributes *AND* via the low-level "object.__dict__" dictionary. Property
    # objects are *NOT* accessible as instance attributes, for obvious reasons.
    assert is_func_property(Class.instance_property) is True

    # Assert this tester rejects *ALL* other callables.
    for some_callable in CALLABLES:
        assert is_func_property(some_callable) is False

# ....................{ TESTS ~ descriptor : staticmethod  }....................
def test_die_unless_func_staticmethod() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.die_unless_func_staticmethod`
    validator.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.utilfunctest import die_unless_func_staticmethod
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        Class,
    )
    from pytest import raises

    # Assert this validator accepts a static method descriptor.
    #
    # Note that static method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as class or instance
    # attributes, static methods reduce to instances of the standard
    # "beartype.cave.FunctionType" type.
    die_unless_func_staticmethod(Class.__dict__['static_method'])

    # Assert this validator rejects *ALL* other callables.
    for some_callable in CALLABLES:
        with raises(_BeartypeUtilCallableException):
            die_unless_func_staticmethod(some_callable)


def test_is_func_staticmethod() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_staticmethod` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_staticmethod
    from beartype_test.a00_unit.data.data_type import (
        CALLABLES,
        Class,
    )

    # Assert this tester accepts a static method descriptor.
    #
    # Note that static method descriptors are *ONLY* directly accessible via the
    # low-level "object.__dict__" dictionary. When accessed as class or instance
    # attributes, static methods reduce to instances of the standard
    # "beartype.cave.FunctionType" type.
    assert is_func_staticmethod(Class.__dict__['static_method']) is True

    # Assert this tester rejects *ALL* other callables.
    for some_callable in CALLABLES:
        assert is_func_staticmethod(some_callable) is False

# ....................{ TESTS ~ python                     }....................
def test_die_unless_func_python() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.die_unless_func_python` tester.
    '''

    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.utilfunctest import die_unless_func_python
    from beartype_test.a00_unit.data.data_type import CALLABLES_C
    from pytest import raises

    # Assert this validator accepts pure-Python callables.
    die_unless_func_python(lambda: True)

    # Assert this validator rejects *ALL* C-based callables.
    for callable_c in CALLABLES_C:
        with raises(_BeartypeUtilCallableException):
            die_unless_func_python(callable_c)


def test_is_func_python() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_python` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_python

    # Assert this tester accepts pure-Python callables.
    assert is_func_python(lambda: True) is True

    # Assert this tester rejects C-based callables.
    assert is_func_python(iter) is False

# ....................{ TESTS ~ nested : closure           }....................
def test_is_func_closure() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_closure` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_closure
    from beartype_test.a00_unit.data.data_type import (
        closure_factory,
        function,
    )

    # Assert this tester accepts closures.
    assert is_func_closure(closure_factory()) is True

    # Assert this tester rejects non-closure callables.
    assert is_func_closure(function) is False

# ....................{ TESTS ~ nested : wrapper           }....................
def test_is_func_wrapper() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_wrapper` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_wrapper
    from beartype_test.a00_unit.data.data_type import (
        decorator_isomorphic,
        function,
    )

    # Assert this tester accepts wrappers generated by standard decorators.
    assert is_func_wrapper(decorator_isomorphic(function)) is True

    # Assert this tester rejects non-wrapper callables.
    assert is_func_wrapper(function) is False

    # Assert this tester rejects non-wrapper non-callables.
    assert is_func_wrapper(
        'Of mouldering leaves in the waste wilderness:—') is False


def test_is_func_wrapper_isomorphic() -> None:
    '''
    Test the
    :func:`beartype._util.func.utilfunctest.is_func_wrapper_isomorphic` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.func.utilfunctest import is_func_wrapper_isomorphic
    from beartype_test.a00_unit.data.data_type import (
        context_manager_factory,
        decorator_isomorphic,
        decorator_nonisomorphic,
        function,
        wrapper_isomorphic,
    )

    # Assert this tester accepts isomorphic closure wrappers.
    assert is_func_wrapper_isomorphic(context_manager_factory) is True
    assert is_func_wrapper_isomorphic(decorator_isomorphic(function)) is True

    # Assert this tester accepts isomorphic non-closure wrappers.
    assert is_func_wrapper_isomorphic(wrapper_isomorphic) is True

    # Assert this tester rejects non-isomorphic wrappers.
    assert is_func_wrapper_isomorphic(
        decorator_nonisomorphic(function)) is False

    # Assert this tester rejects non-wrapper callables.
    assert is_func_wrapper_isomorphic(function) is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype** :pep:`544` **unit tests.**

This submodule unit tests :pep:`544` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_decor_pep544() -> None:
    '''
    Test :pep:`544` support implemented in the :func:`beartype.beartype`
    decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from abc import abstractmethod
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep3119Exception
    from pytest import raises
    from typing import Protocol, runtime_checkable

    # ....................{ PROTOCOLS                      }....................
    @runtime_checkable
    class Easter1916(Protocol):
        '''
        User-defined runtime protocol declaring arbitrary methods.
        '''

        def i_have_met_them_at_close_of_day(self) -> str:
            return 'Coming with vivid faces'

        @abstractmethod
        def from_counter_or_desk_among_grey(self) -> str: pass


    class Easter1916Structural(object):
        '''
        User-defined class structurally (i.e., implicitly) satisfying *without*
        explicitly subclassing this user-defined protocol.
        '''

        def i_have_met_them_at_close_of_day(self) -> str:
            return 'Eighteenth-century houses.'

        def from_counter_or_desk_among_grey(self) -> str:
            return 'I have passed with a nod of the head'


    class TwoTrees(Protocol):
        '''
        User-defined protocol declaring arbitrary methods, but intentionally
        *not* decorated by the :func:`typing.runtime_checkable` decorator and
        thus unusable at runtime by :mod:`beartype`.
        '''

        def holy_tree(self) -> str:
            return 'Beloved, gaze in thine own heart,'

        @abstractmethod
        def bitter_glass(self) -> str: pass

    # ....................{ FUNCTIONS                      }....................
    # @beartype-decorated callable annotated by this user-defined protocol.
    @beartype
    def or_polite_meaningless_words(lingered_awhile: Easter1916) -> str:
        return (
            lingered_awhile.i_have_met_them_at_close_of_day() +
            lingered_awhile.from_counter_or_desk_among_grey()
        )

    # ....................{ PASS                           }....................
    # Assert this callable returns the expected string when passed this
    # user-defined class structurally satisfying this protocol.
    assert or_polite_meaningless_words(Easter1916Structural()) == (
        'Eighteenth-century houses.'
        'I have passed with a nod of the head'
    )

    # ....................{ FAIL                           }....................
    # @beartype-decorated callable annotated by this user-defined protocol.
    with raises(BeartypeDecorHintPep3119Exception):
        @beartype
        def times_of_old(god_slept: TwoTrees) -> str:
            return god_slept.holy_tree() + god_slept.bitter_glass()

# ....................{ TESTS ~ protocol                   }....................
def test_decor_pep544_hint_subprotocol_elision() -> None:
    '''
    Test that type-checking wrappers generated by the :func:`beartype.beartype`
    decorator for callables annotated by one or more **subprotocols** (i.e.,
    :pep:`544`-compliant subclasses subclassing one or more
    :pep:`544`-compliant superclasses themselves directly subclassing the root
    :class:`typing.Protocol` superclass) are optimized to perform only single
    :func:`isinstance` checks against those subprotocols rather than multiple
    :func:`isinstance` checks against both those subprotocols and one or more
    superclasses of those subprotocols.

    This unit test guards against non-trivial performance regressions that
    previously crippled the third-party :mod:`numerary` package, whose caching
    algorithm requires that :mod:`beartype` be optimized in this way.

    See Also
    ----------
    This relevant issue: https://github.com/beartype/beartype/issues/76
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from abc import abstractmethod
    from beartype import beartype
    from typing import (
        Protocol,
        runtime_checkable,
    )

    # Private metaclass of the root "typing.Protocol" superclass, accessed
    # without violating privacy encapsulation.
    _ProtocolMeta = type(Protocol)

    # ..................{ METACLASSES                        }..................
    class UndesirableProtocolMeta(_ProtocolMeta):
        '''
        Undesirable :class:`typing.Protocol`-compliant metaclass additionally
        recording when classes with this metaclass are passed as the second
        parameter to the :func:`isinstance` builtin, enabling subsequent logic
        to validate this is *not* happening as expected.
        '''

        def __instancecheck__(cls, obj: object) -> bool:
            '''
            Unconditionally return ``False``.

            Additionally, this dunder method unconditionally records that this
            method has been called by setting an instance variable of the
            passed object set only by this method.
            '''

            # Set an instance variable of this object set only by this method.
            obj.is_checked_by_undesirable_protocol_meta = True

            # Print a string to standard output as an additional sanity check.
            print("Save when the eagle brings some hunter's bone,")

            # Unconditionally return false.
            return False


    class DesirableProtocolMeta(UndesirableProtocolMeta):
        '''
        Desirable :class:`typing.Protocol`-compliant metaclass.
        '''

        def __instancecheck__(cls, obj: object) -> bool:
            '''
            Unconditionally return ``True``.
            '''

            # Unconditionally return true.
            return True

    # ..................{ PROTOCOLS                          }..................
    @runtime_checkable
    class PileAroundIt(
        Protocol,
        metaclass=UndesirableProtocolMeta,
    ):
        '''
        Arbitrary protocol whose metaclass is undesirable.
        '''

        @abstractmethod
        def ice_and_rock(self, broad_vales_between: str) -> str:
            pass


    @runtime_checkable
    class OfFrozenFloods(
        PileAroundIt,
        Protocol,  # <-- Subprotocols *MUST* explicitly relist this. /facepalm/
        metaclass=DesirableProtocolMeta,
    ):
        '''
        Arbitrary subprotocol of the protocol declared above whose metaclass
        has been replaced with a more desirable metaclass.
        '''

        @abstractmethod
        def and_wind(self, among_the_accumulated_steeps: str) -> str:
            pass

    # ..................{ STRUCTURAL                         }..................
    class ADesertPeopledBy(object):
        '''
        Arbitrary concrete class structurally satisfying *without* subclassing
        the subprotocol declared above.
        '''

        def and_wind(self, among_the_accumulated_steeps: str) -> str:
            return among_the_accumulated_steeps + 'how hideously'

        def ice_and_rock(self, broad_vales_between: str) -> str:
            return broad_vales_between + 'unfathomable deeps,'

    # Arbitrary instance of this class.
    the_storms_alone = ADesertPeopledBy()

    # ..................{ BEARTYPE                           }..................
    @beartype
    def blue_as_the_overhanging_heaven(that_spread: OfFrozenFloods) -> str:
        '''
        Arbitrary callable annotated by the subprotocol declared above.
        '''

        return that_spread.ice_and_rock('Of frozen floods, ')

    # Assert this callable returns the expected string.
    assert blue_as_the_overhanging_heaven(that_spread=the_storms_alone) == (
        'Of frozen floods, unfathomable deeps,')

    # Critically, assert the instance variable *ONLY* defined by the
    # "UndesirableProtocolMeta" metaclass to remain undefined, implying the
    # DesirableProtocolMeta.__instancecheck__() rather than
    # UndesirableProtocolMeta.__instancecheck__() dunder method to have been
    # implicitly called by the single isinstance() call in the body of the
    # type-checking wrapper generated by @beartype above.
    assert not hasattr(
        the_storms_alone, 'is_checked_by_undesirable_protocol_meta')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **functional validation classes** (i.e., :mod:`beartype`-specific
classes enabling callers to define PEP-compliant validators from arbitrary
caller-defined callables *not* efficiently generating stack-free code).

This private submodule defines the core low-level class hierarchy driving the
entire :mod:`beartype` validation ecosystem.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeValeLambdaWarning,
    BeartypeValeValidationException,
)
from beartype.typing import Protocol
from beartype.vale._is._valeisabc import _BeartypeValidatorFactoryABC
from beartype.vale._core._valecore import BeartypeValidator
from beartype.vale._util._valeutilfunc import die_unless_validator_tester
from beartype.vale._util._valeutiltyping import BeartypeValidatorTester
from beartype._data.hint.datahinttyping import LexicalScope
from beartype._util.func.utilfuncscope import add_func_scope_attr
from beartype._util.text.utiltextrepr import (
    represent_func,
    represent_object,
)

# ....................{ PRIVATE ~ protocols                }....................
class _SupportsBool(Protocol):
    '''
    Fast caching protocol matching any object whose class defines the
    :meth:`__bool__` dunder method.
    '''

    def __bool__(self) -> bool: ...


class _SupportsLen(Protocol):
    '''
    Fast caching protocol matching any object whose class defines the
    :meth:`__len__` dunder method.
    '''

    def __len__(self) -> bool: ...


_BoolLike = (_SupportsBool, _SupportsLen)
'''
:func:`isinstance`-able tuple of fast caching protocols matching any
**bool-like** (i.e., object whose class defines at least one of the
:meth:`__bool__` and/or :meth:`__len__` dunder methods).
'''

# ....................{ PRIVATE ~ subclasses               }....................
class _IsFactory(_BeartypeValidatorFactoryABC):
    '''
    **Beartype callable validator factory** (i.e., class that, when subscripted
    (indexed) by an arbitrary callable returning ``True`` when the object
    passed to that callable satisfies a caller-defined constraint, creates a
    new :class:`BeartypeValidator` object encapsulating that callable suitable
    for subscripting (indexing) :attr:`typing.Annotated` type hints, enforcing
    that constraint on :mod:`beartype`-decorated callable parameters and
    returns annotated by those hints).

    This class validates that callable parameters and returns satisfy the
    arbitrary **callable validator** (i.e., callable whose signature satisfies
    ``collections.abc.Callable[[typing.Any], bool]``) subscripting (indexing)
    this class. Callable validators are caller-defined and may thus validate
    the internal integrity, consistency, and structure of arbitrary objects
    ranging from simple builtin scalars like integers and strings to complex
    data structures defined by third-party packages like NumPy arrays and
    Pandas DataFrames.

    This class creates one new :class:`BeartypeValidator` object for each
    callable validator subscripting (indexing) this class. These objects:

    * Are **PEP-compliant** and thus guaranteed to *never* violate existing or
      future standards.
    * Are **Safely ignorable** by *all* static and runtime type checkers other
      than :mod:`beartype` itself.
    * **Less efficient** than :class:`BeartypeValidator` objects created by
      subscripting every other :mod:`beartype.vale` class. Specifically:

      * Every :class:`BeartypeValidator` object created by subscripting this
        class necessarily calls a callable validator and thus incurs at least
        one additional call stack frame per :mod:`beartype`-decorated callable
        call.
      * Every :class:`BeartypeValidator` object created by subscripting every
        other :mod:`beartype.vale` class directly calls *no* callable and thus
        incurs additional call stack frames only when the active Python
        interpreter internally calls dunder methods (e.g., ``__eq__()``) to
        satisfy their validation constraint.

    Usage
    ----------
    Any :mod:`beartype`-decorated callable parameter or return annotated by a
    :attr:`typing.Annotated` type hint subscripted (indexed) by this class
    subscripted (indexed) by a callable validator (e.g.,
    ``typing.Annotated[{cls}, beartype.vale.Is[lambda obj: {expr}]]`` for any
    class ``{cls}``  and Python expression ``{expr}` evaluating to a boolean)
    validates that parameter or return value to be an instance of that class
    satisfying that callable validator.

    Specifically, callers are expected to (in order):

    #. Annotate a callable parameter or return to be validated with a
       :pep:`593`-compliant :attr:`typing.Annotated` type hint.
    #. Subscript that hint with (in order):

       #. The type expected by that parameter or return.
       #. One or more subscriptions (indexations) of this class, each itself
          subscripted (indexed) by a **callable validator** (i.e., callable
          accepting a single arbitrary object and returning either ``True`` if
          that object satisfies an arbitrary constraint *or* ``False``
          otherwise). If that hint is subscripted by:

          * Only one subscription of this class, that parameter or return
            satisfies that hint when both:

            * That parameter or return is an instance of the expected type.
            * That validator returns ``True`` when passed that parameter or
              return.

          * Two or more subscriptions of this class, that parameter or return
            satisfies that hint when both:

            * That parameter or return is an instance of the expected type.
            * *All* callable validators subscripting *all* subscriptions of
              this class return ``True`` when passed that parameter or return.

          Formally, the signature of each callable validator *must* resemble:

          .. code-block:: python

             def is_object_valid(obj) -> bool:
                 return bool(obj)

          Equivalently, each callable validator *must* satisfy the type hint
          ``collections.abc.Callable[[typing.Any,], bool]``. If not the case,
          an exception is raised. Note that:

          * If that parameter or return is *not* an instance of the expected
            type, **no callable validator is called.** Equivalently, each
            callable validator is called *only* when that parameter or return
            is already an instance of the expected type. Callable validators
            need *not* revalidate that type (e.g., by passing that parameter or
            return and type to the :func:`isinstance` builtin).
          * The name of each callable validator is irrelevant. For convenience,
            most callable validators are defined as nameless lambda functions.

    For example, the following type hint only accepts non-empty strings:

    .. code-block:: python

       Annotated[str, Is[lambda text: bool(text)]]

    :class:`BeartypeValidator` objects also support an expressive
    domain-specific language (DSL) enabling callers to trivially synthesize new
    objects from existing objects with standard Pythonic math operators:

    * **Negation** (i.e., ``not``). Negating an :class:`BeartypeValidator`
      object with the ``~`` operator synthesizes a new
      :class:`BeartypeValidator` object whose validator returns ``True`` only
      when the validator of the original object returns ``False``. For example,
      the following type hint only accepts strings containing *no* periods:

      .. code-block:: python

         Annotated[str, ~Is[lambda text: '.' in text]]

    * **Conjunction** (i.e., ``and``). Conjunctively combining two or more
      :class:`BeartypeValidator` objects with the ``&`` operator synthesizes a
      new :class:`BeartypeValidator` object whose validator returns ``True``
      only when all data validators of the original objects return ``True``.
      For example, the following type hint only accepts non-empty strings
      containing *no* periods:

      .. code-block:: python

         Annotated[str, (
              Is[lambda text: bool(text)] &
             ~Is[lambda text: '.' in text]
         )]

    * **Disjunction** (i.e., ``or``). Disjunctively combining two or more
      :class:`BeartypeValidator` objects with the ``|`` operator synthesizes a
      new :class:`BeartypeValidator` object whose validator returns ``True``
      only when at least one validator of the original objects returns
      ``True``. For example, the following type hint accepts both empty strings
      *and* non-empty strings containing at least one period:

      .. code-block:: python

         Annotated[str, (
             ~Is[lambda text: bool(text)] |
              Is[lambda text: '.' in text]
         )]

    See also the **Examples** subsection below.

    Caveats
    ----------
    **This class is currently only supported by the** :func:`beartype.beartype`
    **decorator.** All other static and runtime type checkers silently ignore
    subscriptions of this class subscripting :attr:`typing.Annotated` type
    hints.

    **This class incurs a minor time performance penalty at call time.**
    Specifically, each type hint of a :mod:`beartype`-decorated callable
    subscripted by a subscription of this class adds one additional stack frame
    to each call of that callable. While negligible (in the average case), this
    cost can become non-negligible when compounded across multiple type hints
    annotating a frequently called :mod:`beartype`-decorated callable --
    especially when those type hints are subscripted by multiple subscriptions
    of this class at different nesting levels.

    **This class prohibits instantiation.** This class is *only* intended to be
    subscripted. Attempting to instantiate this class into an object will raise
    an :exc:`.BeartypeValeSubscriptionException` exception.

    Examples
    ----------
    .. code-block:: python

       # Import the requisite machinery.
       >>> from beartype import beartype
       >>> from beartype.vale import Is
       >>> from typing import Annotated

       # Validator matching only strings with lengths ranging [4, 40].
       >>> IsRangy = Is[lambda text: 4 <= len(text) <= 40]

       # Validator matching only unquoted strings.
       >>> IsUnquoted = Is[lambda text:
       ...     text.count('"') < 2 and text.count("'") < 2]

       # Type hint matching only unquoted strings.
       >>> UnquotedString = Annotated[str, IsUnquoted]

       # Type hint matching only quoted strings.
       >>> QuotedString = Annotated[str, ~IsUnquoted]

       # Type hint matching only unquoted strings with lengths ranging [4, 40].
       >>> UnquotedRangyString = Annotated[str, IsUnquoted & IsRangy]

       # Annotate callables by those type hints.
       >>> @beartype
       ... def doublequote_text(text: UnquotedString) -> QuotedString:
       ...     """
       ...     Double-quote the passed unquoted string.
       ...     """
       ...     return f'"{text}"'  # The best things in life are one-liners.
       >>> @beartype
       ... def singlequote_prefix(text: UnquotedRangyString) -> QuotedString:
       ...     """
       ...     Single-quote the prefix spanning characters ``[0, 3]`` of the
       ...     passed unquoted string with length ranging ``[4, 40]``.
       ...     """
       ...     return f"'{text[:3]}'"  # "Guaranteed to work," says @beartype.

       # Call those callables with parameters satisfying those validators.
       >>> doublequote_text("You know anything about nuclear fusion?")
       "You know anything about nuclear fusion?"
       >>> singlequote_prefix("Not now, I'm too tired. Maybe later.")
       'Not'

       # Call those callables with parameters not satisfying those validators.
       >>> doublequote_text('''"Everybody relax, I'm here."''')
       beartype.roar._roarexc.BeartypeCallHintParamViolation: @beartyped
       doublequote_text() parameter text='"Everybody relax, I\'m here."'
       violates type hint typing.Annotated[str, Is[lambda text: text.count('"')
       < 2 and text.count("'") < 2]], as value '"Everybody relax, I\'m here."'
       violates validator Is[lambda text: text.count('"') < 2 and
       text.count("'") < 2].
    '''

    # ..................{ DUNDERS                            }..................
    def __getitem__(  # type: ignore[override]
        self, is_valid: BeartypeValidatorTester) -> BeartypeValidator:
        '''
        Create and return a new beartype validator from the passed **validator
        callable** (i.e., caller-defined callable accepting a single arbitrary
        object and returning either ``True`` if that object satisfies an
        arbitrary constraint *or* ``False`` otherwise), suitable for
        subscripting :pep:`593`-compliant :attr:`typing.Annotated` type hints.

        This method is intentionally *not* memoized, as this method is usually
        subscripted only by subscription-specific lambda functions uniquely
        defined for each subscription of this class.

        Parameters
        ----------
        is_valid : Callable[[Any,], bool]
            Validator callable to validate parameters and returns against.

        Returns
        ----------
        BeartypeValidator
            New object encapsulating this validator callable.

        Raises
        ----------
        BeartypeValeSubscriptionException
            If either:

            * This class was subscripted by two or more arguments.
            * This class was subscripted by one argument that either:

              * Is *not* callable.
              * Is a C-based rather than pure-Python callable.
              * Is a pure-Python callable accepting two or more arguments.

        See Also
        ----------
        :class:`_IsAttrFactory`
            Usage instructions.
        '''

        # ..................{ VALIDATE                       }..................
        # If this class was subscripted by either no arguments *OR* two or more
        # arguments, raise an exception.
        self._die_unless_getitem_args_1(is_valid)
        # Else, this class was subscripted by exactly one argument.

        # If that callable is *NOT* a validator tester, raise an exception.
        die_unless_validator_tester(is_valid)
        # Else, that callable is a validator tester.

        # Lambda function dynamically generating the machine-readable
        # representation of this validator, deferred due to the computational
        # expense of accurately retrieving the source code for this validator
        # (especially when this validator is itself a lambda function).
        get_repr = lambda: (
            f'{self._basename}['
            f'{represent_func(func=is_valid, warning_cls=BeartypeValeLambdaWarning)}'
            f']'
        )

        # ..................{ CLOSURE                        }..................
        #FIXME: Unit test edge cases extensively, please.
        def _is_valid_bool(obj: object) -> bool:
            '''
            :data:`True` only if the passed object satisfies the caller-defined
            validation callable subscripting this :attr:`beartype.vale.Is`
            validator factory.

            This closure wraps that possibly unsafe callable with an implicit
            type cast, guaranteeing that either:

            * If that callable returns a boolean, this closure returns that
              boolean as is.
            * If that callable returns a non-boolean object, either:

              * If that non-boolean is implicitly convertible into a boolean
                (i.e., if passing that non-boolean to the :class:`bool` type
                succeeds *without* raising an exception), this closure coerces
                that non-boolean into a boolean and returns that boolean.
              * Else, this closure raises a human-readable exception.

            This closure is principally intended to massage non-standard
            validation callables defined by popular third-party packages like
            NumPy, which commonly return non-boolean objects that are implicitly
            convertible into boolean objects: e.g.,

            .. code-block::

               >>> import numpy as np
               >>> matrix = np.array([[2, 1], [1, 2]])
               >>> is_all = np.all(matrix > 0))
               >>> type(is_all)
               <class 'numpy.bool_'>
               >>> is_all
               True  # <-- y u lie, numpy
               >>> bool(is_all)
               True

            Caveats
            ----------
            **This closure is comparatively slower than the passed callable.**
            This closure should *never* be called directly from code snippets
            embedded in wrapper functions dynamically generated by the
            :func:`beartype.beartype` decorator. This closure should *only* be
            called indirectly by exception-handling functionality performed by
            those wrapper functions in the event of a type-checking violation,
            at which time efficiency is no longer a driving force.

            This implies that wrapper functions dynamically generated by the
            :func:`beartype.beartype` decorator *could* implicitly coerce
            non-boolean objects returned by the passed callable into the
            ``True`` singleton. Although non-ideal, debugging such concerns is
            squarely the user's concern; attempting to safeguard users from
            semantic issues like this would destroy runtime performance for *no*
            tangible gain in the general case. The best :mod:`beartype` can (and
            should) do is defer validation until a type-checking violation.

            Parameters
            ----------
            obj : object
                Object to be validated by that validation callable.

            Returns
            ----------
            bool
                :data:`True` only if that object satisfies that validation
                callable.

            Raises
            ----------
            BeartypeValeValidationException
                If that validation callable returns a **non-bool-like**, where
                "non-bool-like" is any object that:

                * Is *not* a **boolean** (i.e., :class:`bool` instance).
                * Is *not* **implicitly convertible** into a boolean (i.e., is
                  an object whose class defines neither the :meth:`__bool__` nor
                  :meth:`__len__` dunder methods).

    * Subscript the :attr:`beartype.vale.Is` factory by a **non-bool-like
      validator** (i.e., tester function returning an object that is neither a
      :class:`bool` *nor* implicitly convertible into a :class:`bool`).
            '''

            # Object returned by validating this object against that callable.
            is_obj_valid = is_valid(obj)

            # If that object is a boolean, return that object as is.
            if isinstance(is_obj_valid, bool):
                return is_obj_valid
            # Else, that object is *NOT* a boolean.

            # "True" *ONLY* if that object is a bool-like (i.e., object whose
            # class defines the __bool__() and/or __len__() dunder methods).
            #
            # Note that we intentionally avoid the Easier to Ask for Permission
            # than Forgiveness (EAFP) approach typically favoured by the Python
            # community for coercing types. Namely, we avoid doing this:
            #    # Attempt to coerce this boolean into a non-boolean.
            #    try:
            #        is_obj_valid_bool = bool(is_obj_valid)
            #    except Exception as exception:
            #        raise SomeBeartypeException(...) from exception
            #
            # Why? Because the bool() constructor is overly permissive to the
            # point of being *FRANKLY BROKEN.* Why? Because that constructor
            # *NEVER* raises an exception (unless the class of that object
            # defines a __bool__() dunder method raising an exception). Why?
            # Because that constructor implicitly coerces *ALL* objects whose
            # classes define *NO* __bool__() dunder method to "True" except for
            # the following, which the bool() constructor explicitly detects
            # and hard-codes to be coerced to "False":
            # * The "None" singleton.
            # * The "False" singleton.
            # * Numeric 0 across all numeric types, including:
            #   * Integer 0.
            #   * Floating-point 0.0.
            # * Empty containers across all container types, including:
            #   * The empty tuple singleton (i.e., "()").
            #   * The empty string singleton (i.e., "''").
            #   * Any empty list (e.g., "[]").
            #
            # The proof is in the gelatinous spaghetti code:
            #     >>> class OhMyGods(object): pass
            #     >>> bool(OhMyGods())
            #     True  # <-- WHAT THE HECK IS THIS, GUIDO. SRSLY, BRO. SRSLY.
            #
            # This is, of course, unbelievable. This is, of course, all true.
            # What is this, Guido? Visual Basic in my Python? *facepalm*
            #
            # Note also that there are several means of testing for booliness.
            # The obvious approach of calling getattr() is also the slowest,
            # because getattr() internally performs the EAFP approach and
            # exception handling in Python is known to be an obvious bottleneck.
            # Ergo, we intentionally avoid doing this:
            #     is_obj_valid_bool_method = getattr(is_obj_valid, '__bool__', None)
            #
            # Ideally, we would instead defer to a beartype-specific fast
            # caching protocol that also internally performs a similar getattr()
            # call wrapped within caching logic that amortizes the cost of that
            # call across all isinstance() calls passed an object of that same
            # type. Since there exists *NO* standard "SupportsBool" protocol,
            # we would then trivially define our own like so:
            #     from beartype.typing import Protocol
            #     class SupportsBool(Protocol):
            #         def __bool__(self) -> bool: ...
            #
            # Surprisingly, that fails. Why? Because the bool() constructor
            # internally coerces objects into booleans like so:
            # * If the passed object defines the __bool__() dunder method, that
            #   constructor defers to that method.
            # * Else if the passed object defines the __len__() dunder method,
            #   that constructor defers to that method.
            # * Else if the passed object is one of several hard-coded objects
            #   evaluating to "False", that constructor returns "False".
            # * Else, that constructor returns "True".
            #
            # To handle the first two cases, we instead:
            # * Define both our own "SupportsBool" and "SupportsLen" protocols.
            # * Decide whether that object is bool-like by deferring to those
            #   protocols.
            is_obj_valid_boollike = isinstance(is_obj_valid, _BoolLike)  # pyright: ignore

            # If that object is *NOT* bool-like, raise an exception.
            if not is_obj_valid_boollike:
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # CAUTION: Synchronize with the exception raised below, please.
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                raise BeartypeValeValidationException(
                    f'Validator {get_repr()} '
                    f'return value {repr(is_obj_valid)} not bool-like '
                    f'(i.e., instance of neither "bool" nor '
                    f'class defining __bool__() or __len__() dunder methods) '
                    f'for subject object:\n{represent_object(obj)}'
                )
            # Else, that object is bool-like.

            # Boolean coerced from this non-boolean via the __bool__() or
            # __len__() dunder methods declared by the type of this non-boolean,
            # initialized to "False" for safety.
            is_obj_valid_bool = False

            # Attempt to perform this coercion.
            try:
                is_obj_valid_bool = bool(is_obj_valid)
            # If whichever of the __bool__() or __len__() dunder methods is
            # called by the above bool() constructor raises an exception, wrap
            # that exception in a higher-level @beartype exception.
            #
            # Note that this is *NOT* simply an uncommon edge case. In
            # particular, the Pandas "DataFrame" type defines a __bool__()
            # dunder method that unconditionally raises an exception. *facepalm*
            except Exception as exception:
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # CAUTION: Synchronize with the exception raised above, please.
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                raise BeartypeValeValidationException(
                    f'Validator {get_repr()} '
                    f'return value {repr(is_obj_valid)} erroneously bool-like '
                    f'(i.e., instance of class defining __bool__() or __len__() '
                    f'dunder methods raising unexpected exception) '
                    f'for subject object:\n{represent_object(obj)}'
                ) from exception

            # Return this boolean.
            return is_obj_valid_bool

        # ..................{ VALIDATOR                      }..................
        # Dictionary mapping from the name to value of each local attribute
        # referenced in the "is_valid_code" snippet defined below.
        is_valid_code_locals: LexicalScope = {}

        # Name of a new parameter added to the signature of each
        # @beartype-decorated wrapper function whose value is this validator,
        # enabling this validator to be called directly in the body of those
        # functions *WITHOUT* imposing additional stack frames.
        is_valid_attr_name = add_func_scope_attr(
            attr=_is_valid_bool, func_scope=is_valid_code_locals)

        # One one-liner to rule them all and in "pdb" bind them.
        return BeartypeValidator(
            is_valid=_is_valid_bool,
            # Python code snippet calling this validator (via this new
            # parameter), passed an object to be interpolated into this snippet
            # by downstream logic.
            is_valid_code=f'{is_valid_attr_name}({{obj}})',
            is_valid_code_locals=is_valid_code_locals,
            get_repr=get_repr,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable globals** (i.e., global constants describing various
well-known functions and methods).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................

# ....................{ SETS                               }....................
METHOD_NAMES_DUNDER_BINARY = frozenset((
    '__add__',
    '__and__',
    '__cmp__',
    '__divmod__',
    '__div__',
    '__eq__',
    '__floordiv__',
    '__ge__',
    '__gt__',
    '__iadd__',
    '__iand__',
    '__idiv__',
    '__ifloordiv__',
    '__ilshift__',
    '__imatmul__',
    '__imod__',
    '__imul__',
    '__ior__',
    '__ipow__',
    '__irshift__',
    '__isub__',
    '__itruediv__',
    '__ixor__',
    '__le__',
    '__lshift__',
    '__lt__',
    '__matmul__',
    '__mod__',
    '__mul__',
    '__ne__',
    '__or__',
    '__pow__',
    '__radd__',
    '__rand__',
    '__rdiv__',
    '__rfloordiv__',
    '__rlshift__',
    '__rmatmul__',
    '__rmod__',
    '__rmul__',
    '__ror__',
    '__rpow__',
    '__rrshift__',
    '__rshift__',
    '__rsub__',
    '__rtruediv__',
    '__rxor__',
    '__sub__',
    '__truediv__',
    '__xor__',
))
'''
Frozen set of the unqualified names of all **binary dunder methods** (i.e.,
methods whose names are both prefixed and suffixed by ``__``, which the active
Python interpreter implicitly calls to perform binary operations on instances
whose first operands are instances of the classes declaring those methods).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype slow cave** (i.e., private subset of the public :mod:`beartype.cave`
subpackage profiled to *not* be efficiently importable at :mod:`beartype`
startup and thus *not* safely importable throughout the internal
:mod:`beartype` codebase).

This submodule currently imports from expensive third-party packages on
importation (e.g., :mod:`numpy`) despite :mod:`beartype` itself *never*
requiring those imports. Until resolved, that subpackage is considered tainted.
'''

# ....................{ TODO                               }....................
#FIXME: Excise this submodule away, please. This submodule was a horrendous idea
#and has plagued the entire "beartype.cave" subpackage with unnecessary slowdown
#at import time. It's simply time for this to go, please.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid polluting the public module namespace, external attributes
# should be locally imported at module scope *ONLY* under alternate private
# names (e.g., "from argparse import ArgumentParser as _ArgumentParser" rather
# than merely "from argparse import ArgumentParser").
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from argparse import (
    ArgumentParser,
    _SubParsersAction,
)
from weakref import (
    ProxyTypes,
    ref,
)

# ....................{ TYPES ~ lib                        }....................
# Types conditionally dependent upon the importability of third-party
# dependencies. These types are subsequently redefined by try-except blocks
# below and initially default to "UnavailableType" for simple types.

# ....................{ TYPES ~ stdlib : argparse          }....................
ArgParserType = ArgumentParser
'''
Type of argument parsers parsing all command-line arguments for either
top-level commands *or* subcommands of those commands.
'''


ArgSubparsersType = _SubParsersAction
'''
Type of argument subparser containers parsing subcommands for parent argument
parsers parsing either top-level commands *or* subcommands of those commands.
'''

# ....................{ TYPES ~ stdlib : weakref           }....................
WeakRefCType = ref
'''
Type of all **unproxied weak references** (i.e., callable objects yielding
strong references to their referred objects when called).

This type matches both the C-based :class:`weakref.ref` class *and* the
pure-Python :class:`weakref.WeakMethod` class, which subclasses the former.
'''
# ....................{ TUPLES ~ stdlib : weakref          }....................
WeakRefProxyCTypes = ProxyTypes
'''
Tuple of all **C-based weak reference proxy classes** (i.e., classes
implemented in low-level C whose instances are weak references to other
instances masquerading as those instances).

This tuple contains classes matching both callable and uncallable weak
reference proxies.
'''



#!/usr/bin/env python3
#--------------------( LICENSE                             )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Decidedly Object-Oriented Runtime-checking (DOOR) callable type hint classes**
(i.e., :class:`beartype.door.TypeHint` subclasses implementing support for
:pep:`484`- and :pep:`585`-compliant ``Tuple[...]`` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsub import _TypeHintSubscripted
from beartype.door._cls.doorsuper import TypeHint
from beartype.typing import Any

# ....................{ SUBCLASSES                         }....................
# FIXME: Document all public and private attributes of this class, please.
class _TupleTypeHint(_TypeHintSubscripted):
    '''
    **Tuple type hint wrapper** (i.e., high-level object encapsulating a
    low-level :pep:`484`- or :pep:`585`-compliant ``Tuple[...]`` type hint).

    Attributes (Private)
    --------
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, hint: object) -> None:

        #FIXME: Actually, it might be preferable to define two distinct
        #subclasses:
        #* "class _TypeHintTupleVariadic(_TypeHintSequence)", handling variadic
        #  tuple type hints of the form "Tuple[T, ...]".
        #* "class _TypeHintTupleFixed(_TypeHintSubscripted)", handling
        #  fixed-length tuple type hints.
        #
        #Why? Because variadic and fixed-length tuple type hints have *NOTHING*
        #semantically to do with one another. All they share is the common
        #prefix "Tuple". Aside from that, everything is dissimilar. Indeed,
        #most of the logic below (especially _is_subhint_branch(), which is kinda
        #cray-cray) would strongly benefit from separating this class into two
        #subclasses.
        #
        #Note that implementing this division will probably require generalizing
        #the TypeHint.__new__() method to support this division.

        # Initialize all subclass-specific instance variables for safety.
        self._is_variable_length: bool = False

        #FIXME: Non-ideal. The superclass __len__() method already returns 0 as
        #expected for "Tuple[()]" type hints. Excise us up!
        self._is_empty_tuple: bool = False

        # Initialize the superclass with all passed parameters.
        super().__init__(hint)


    def _make_args(self) -> tuple:

        # Tuple of the zero or more low-level child type hints subscripting
        # (indexing) the low-level parent type hint wrapped by this wrapper.
        args = super()._make_args()

        # Validate these child hints. Specifically, remove any
        # PEP-noncompliant child hints from this tuple and set associated flags.
        #
        # e.g. `Tuple` without any arguments
        # This may be unreachable (since a bare Tuple will go to ClassTypeHint),
        # but it's here for completeness and safety.
        if len(args) == 0:  # pragma: no cover
            args = (Any,)
            self._is_variable_length = True
        #FIXME: This is non-portable and thus basically broken. Ideally, empty
        #tuples should instead be detected by explicitly calling the
        #is_hint_pep484585_tuple_empty() tester.
        elif len(args) == 1 and args[0] == ():
            args = ()
            self._is_empty_tuple = True
        elif len(args) == 2 and args[1] is Ellipsis:
            args = (args[0],)
            self._is_variable_length = True

        # Return these child hints.
        return args

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _is_args_ignorable(self) -> bool:

        #FIXME: Actually, pretty sure this only returns true if this hint is
        #"Tuple[Any, ...]". *shrug*
        # Return true only if this hint is either "Tuple[Any, ...]" or the
        # unsubscripted "Tuple" type hint factory.
        return (
            self._is_variable_length and
            bool(self) and
            self[0].is_ignorable
        )

    # ..................{ PRIVATE ~ testers                  }..................
    def _is_subhint_branch(self, branch: TypeHint) -> bool:

        if branch._is_args_ignorable:
            return issubclass(self._origin, branch._origin)
        elif not isinstance(branch, _TupleTypeHint):
            return False
        elif self._is_args_ignorable:
            return False
        elif branch._is_empty_tuple:
            return self._is_empty_tuple
        elif branch._is_variable_length:
            that_hint = branch._args_wrapped_tuple[0]

            if self._is_variable_length:
                return that_hint.is_subhint(self._args_wrapped_tuple[0])

            return all(
                this_child.is_subhint(that_hint)
                for this_child in self._args_wrapped_tuple
            )
        elif self._is_variable_length:
            return (
                branch._is_variable_length and
                self._args_wrapped_tuple[0].is_subhint(
                    branch._args_wrapped_tuple[0])
            )
        elif len(self._args) != len(branch._args):
            return False

        return all(
            this_child <= that_child
            for this_child, that_child in zip(
                self._args_wrapped_tuple, branch._args_wrapped_tuple
            )
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`589`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.hint.datahinttyping import MappingStrToAny
from beartype._util.cls.utilclstest import is_type_subclass
from beartype._util.py.utilpyversion import IS_PYTHON_3_8

# ....................{ TESTERS                            }....................
# The implementation of the "typing.TypedDict" attribute substantially varies
# across Python interpreter *AND* "typing" implementation. Specifically:
# * The "typing.TypedDict" attribute under Python >= 3.9 is *NOT* actually a
#   superclass but instead a factory function masquerading as a superclass by
#   setting the subversive "__mro_entries__" dunder attribute to a tuple
#   containing a private "typing._TypedDict" superclass. This superclass
#   necessarily defines the three requisite dunder attributes.
# * The "typing_extensions.TypedDict" attribute under Python < 3.8 is actually
#   a superclass also necessarily defining the three requisite dunder
#   attributes.
# * The "typing.TypedDict" attribute under *ONLY* Python 3.8 is also actually
#   a superclass that *ONLY* defines the requisite "__annotations__" dunder
#   attribute. The two remaining dunder attributes are only conditionally
#   defined and thus *CANNOT* be unconditionally assumed to exist.
# In all three cases, passing the passed hint and that superclass to the
# issubclass() builtin fails, as the metaclass of that superclass prohibits
# issubclass() checks. I am throwing up in my mouth as I write this.
#
# Unfortunately, all of the above complications are further complicated by the
# "dict" type under Python >= 3.10. For unknown reasons, Python >= 3.10 adds
# spurious "__annotations__" dunder attributes to "dict" subclasses -- even if
# those subclasses annotate *NO* class or instance variables. While a likely
# bug, we have little choice but to at least temporarily support this insanity.
def is_hint_pep589(hint: object) -> bool:
    '''
    ``True`` only if the passed object is a :pep:`589`-compliant **typed
    dictionary** (i.e., :class:`typing.TypedDict` subclass).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator). Although the implementation
    inefficiently performs three calls to the :func:`hasattr` builtin (which
    inefficiently calls the :func:`getattr` builtin and catches the
    :exc:`AttributeError` exception to detect false cases), callers are
    expected to instead (in order):

    #. Call the memoized
       :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_sign_or_none`
       getter, which internally calls this unmemoized tester.
    #. Compare the object returned by that getter against the
       :attr:`beartype._util.data.hint.pep.sign.datapepsigns.HintSignTypedDict`
       sign.

    Parameters
    ----------
    hint : object
        Object to be tested.

    Returns
    ----------
    bool
        ``True`` only if this object is a typed dictionary.
    '''

    # If this hint is *NOT* a "dict" subclass, this hint *CANNOT* be a typed
    # dictionary. By definition, typed dictionaries are "dict" subclasses.
    #
    # Note that PEP 589 actually lies about the type of typed dictionaries:
    #     Methods are not allowed, since the runtime type of a TypedDict object
    #     will always be just dict (it is never a subclass of dict).
    #
    # This is *ABSOLUTELY* untrue. PEP 589 authors plainly forgot to implement
    # this constraint. Contrary to the above:
    # * All typed dictionaries are subclasses of "dict".
    # * The type of typed dictionaries is the private "typing._TypedDictMeta"
    #   metaclass across all Python versions (as of this comment).
    #
    # This is where we generously and repeatedly facepalm ourselves.
    if not is_type_subclass(hint, dict):
        return False
    # Else, this hint is a "dict" subclass and thus *MIGHT* be a typed
    # dictionary.

    # Return true *ONLY* if this "dict" subclass defines all three dunder
    # attributes guaranteed to be defined by all typed dictionaries. Although
    # slow, this is still faster than the MRO-based approach delineated above.
    #
    # Note that *ONLY* the Python 3.8-specific implementation of
    # "typing.TypedDict" fails to unconditionally define the
    # "__required_keys__" and "__optional_keys__" dunder attributes. Ergo, if
    # the active Python interpreter targets exactly Python 3.8, we relax this
    # test to *ONLY* test for the "__annotations__" dunder attribute.
    # Specifically, we return true only if...
    #
    # Technically, this test can also be performed by inefficiently violating
    # privacy encapsulation. Specifically, this test could perform an O(k) walk
    # up the class inheritance tree of the passed class (for k the number of
    # superclasses of that class), iteratively comparing each such superclass
    # for against the "typing.TypeDict" superclass. That is, this tester could
    # crazily reimplement the issubclass() builtin in pure-Python. Since the
    # implementation of typed dictionaries varies substantially across Python
    # versions, doing so would require version-specific tests in addition to
    # unsafely violating privacy encapsulation and inefficiently violating
    # constant-time guarantees.
    #
    # Technically, the current implementation of this test is susceptible to
    # false positives in unlikely edge cases. Specifically, this test queries
    # for dunder attributes and thus erroneously returns true for user-defined
    # "dict" subclasses *NOT* subclassing the "typing.TypedDict" superclass but
    # nonetheless declaring the same dunder attributes declared by that
    # superclass. Since the likelihood of any user-defined "dict" subclass
    # accidentally defining these attributes is vanishingly small *AND* since
    # "typing.TypedDict" usage is largely discouraged in the typing community,
    # this error is unlikely to meaningfully arise in real-world use cases.
    # Ergo, it is preferable to implement this test portably, safely, and
    # efficiently rather than accommodate this error.
    #
    # In short, the current approach of is strongly preferable.
    return (
        # This "dict" subclass defines these "TypedDict" attributes *AND*...
        hasattr(hint, '__annotations__') and
        hasattr(hint, '__total__') and
        # Either...
        (
            # The active Python interpreter targets exactly Python 3.8 and
            # thus fails to unconditionally define the remaining attributes
            # *OR*...
            IS_PYTHON_3_8 or
            # The active Python interpreter targets any other Python version
            # and thus unconditionally defines the remaining attributes.
            (
                hasattr(hint, '__required_keys__') and
                hasattr(hint, '__optional_keys__')
            )
        )
    )

# ....................{ REDUCERS                           }....................
#FIXME: Remove *AFTER* deeply type-checking typed dictionaries. For now,
#shallowly type-checking such hints by reduction to untyped dictionaries
#remains the sanest temporary work-around.
def reduce_hint_pep589(
    hint: object, exception_prefix: str, *args, **kwargs) -> object:
    '''
    Reduce the passed :pep:`589`-compliant **typed dictionary** (i.e.,
    :class:`typing.TypedDict` subclass) to a lower-level type hint currently
    supported by :mod:`beartype`.

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : object
        Typed dictionary to be reduced.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message.

    All remaining passed arguments are silently ignored.

    Returns
    ----------
    object
        Lower-level type hint currently supported by :mod:`beartype`.
    '''

    # Silently ignore all child type hints annotating this dictionary by
    # reducing this hint to a "Mapping" type hint. Yes, "Mapping" rather than
    # "dict". By PEP 589 edict:
    #     First, any TypedDict type is consistent with Mapping[str, object].
    return MappingStrToAny



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **sign classes** (i.e., classes whose instances uniquely
identifying PEP-compliant type hints in a safe, non-deprecated manner
regardless of the Python version targeted by the active Python interpreter).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Union

# ....................{ CLASSES                            }....................
class HintSign(object):
    '''
    **Sign** (i.e., object uniquely identifying PEP-compliant type hints in a
    safe, non-deprecated manner regardless of the Python version targeted by
    the active Python interpreter).

    Attributes
    ----------
    name : str
        Uniqualified name of the :mod:`typing` attribute uniquely identified by
        this sign (e.g., ``Literal`` for :pep:`586`-compliant type hints).
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = ('name',)

    # ..................{ DUNDERS                            }..................
    def __init__(self, name: str) -> None:
        '''
        Initialize this sign.

        Parameters
        ----------
        name : str
            Uniqualified name of the :mod:`typing` attribute uniquely
            identified by this sign (e.g., ``Literal`` for :pep:`586`-compliant
            type hints).
        '''
        assert isinstance(name, str), f'{repr(name)} not string.'

        # Classify all passed parameters.
        self.name = name


    def __repr__(self) -> str:
        '''
        Machine-readable representation of this sign.
        '''

        return f"HintSign('{self.name}')"


    def __str__(self) -> str:
        '''
        Human-readable stringification of this sign.
        '''

        return f'"HintSign{self.name}"'

# ....................{ HINTS                              }....................
HintSignOrType = Union[HintSign, type]
'''
PEP-compliant type hint matching either a **sign** (i.e., object uniquely
identifying PEP-compliant type hints in a safe, non-deprecated manner
regardless of the Python version targeted by the active Python interpreter) or
**isinstanceable class** (i.e., class safely passable as the second argument to
the :func:`isinstance` builtin).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`586`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep586Exception
from beartype.typing import Any
from beartype._cave._cavefast import EnumMemberType, NoneType
from beartype._data.hint.datahinttyping import TypeException
from beartype._data.hint.pep.sign.datapepsigns import HintSignLiteral
from beartype._util.text.utiltextjoin import join_delimited_disjunction_types

# ....................{ CONSTANTS                          }....................
_LITERAL_ARG_TYPES = (bool, bytes, int, str, EnumMemberType, NoneType)
'''
Tuple of all types of objects permissible as arguments subscripting the
:pep:`586`-compliant :attr:`typing.Literal` singleton.

These types are explicitly listed by :pep:`586` as follows:

    Literal may be parameterized with literal ints, byte and unicode strings,
    bools, Enum values and None.
'''

# ....................{ VALIDATORS                         }....................
def die_unless_hint_pep586(
    # Mandatory parameters.
    hint: Any,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep586Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is a
    :pep:`586`-compliant type hint (i.e., subscription of either the
    :attr:`typing.Literal` or :attr:`typing_extensions.Literal` type hint
    factories).

    Ideally, the :attr:`typing.Literal` singleton would internally validate the
    literal objects subscripting that singleton at subscription time (i.e., in
    the body of the ``__class_getitem__()`` dunder method). Whereas *all* other
    :mod:`typing` attributes do just that, :attr:`typing.Literal` permissively
    accepts all possible arguments like a post-modern philosopher hopped up on
    too much tenure. For inexplicable reasons, :pep:`586` explicitly requires
    third-party type checkers (that's us) to validate these hints rather than
    standardizing that validation in the :mod:`typing` module. Weep, Guido!

    Caveats
    ----------
    **This function is slow** and should thus be called only once per
    visitation of a :pep:`586`-compliant type hint. Specifically, this function
    is O(n) for n the number of arguments subscripting this hint.

    Parameters
    ----------
    hint : object
        Object to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep586Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Raises
    ----------
    :exc:`exception_cls`
        If this object either:

        * Is *not* a subscription of either the :attr:`typing.Literal` or
          :attr:`typing_extensions.Literal` type hint factories.
        * Subscripts either factory with zero arguments via the empty tuple,
          which these factories sadly fails to guard against.
        * Subscripts either factory with one or more arguments that are *not*
          **valid literals**, defined as the set of all:

          * Booleans.
          * Byte strings.
          * Integers.
          * Unicode strings.
          * :class:`enum.Enum` members.
          * The ``None`` singleton.
    '''

    # Tuple of zero or more literal objects subscripting this hint.
    hint_literals = get_hint_pep586_literals(
        hint=hint,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If the caller maliciously subscripted this hint by the empty tuple and
    # thus *NO* arguments, raise an exception. Ideally, the "typing.Literal"
    # singleton would guard against this itself. It does not; thus, we do.
    if not hint_literals:
        raise exception_cls(
            f'{exception_prefix}PEP 586 type hint {repr(hint)} '
            f'subscripted by empty tuple.'
        )

    # If any argument subscripting this hint is *NOT* a valid literal...
    #
    # Sadly, despite PEP 586 imposing strict restrictions on the types of
    # objects permissible as arguments subscripting the "typing.Literal"
    # singleton, PEP 586 explicitly offloads the odious chore of enforcing
    # those restrictions onto third-party type checkers by intentionally
    # implementing that singleton to permissively accept *ALL* possible
    # objects when subscripted:
    #     Although the set of parameters Literal[...] may contain at type
    #     check time is very small, the actual implementation of
    #     typing.Literal will not perform any checks at runtime.
    if any(
        not isinstance(hint_literal, _LITERAL_ARG_TYPES)
        for hint_literal in hint_literals
    # Then raise a human-readable exception describing this invalidity.
    ):
        # For each argument subscripting this hint...
        for hint_literal_index, hint_literal in enumerate(hint_literals):
            # If this argument is invalid as a literal argument...
            if not isinstance(hint_literal, _LITERAL_ARG_TYPES):
                # Human-readable concatenation of the types of all valid
                # literal arguments, delimited by commas and/or "or".
                hint_literal_types = join_delimited_disjunction_types(
                    _LITERAL_ARG_TYPES)

                # Raise an exception.
                raise exception_cls(
                    f'{exception_prefix}PEP 586 type hint {repr(hint)} '
                    f'argument {hint_literal_index} '
                    f'{repr(hint_literal)} not {hint_literal_types}.'
                )

# ....................{ GETTERS                            }....................
def get_hint_pep586_literals(
    # Mandatory parameters.
    hint: Any,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep586Exception,
    exception_prefix: str = '',
) -> tuple:
    '''
    Tuple of zero or more literal objects subscripting the passed
    :pep:`586`-compliant type hint (i.e., subscription of either the
    :attr:`typing.Literal` or :attr:`typing_extensions.Literal` type hint
    factories).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    ----------
    **This low-level getter performs no validation of the contents of this
    tuple.** Consider calling the high-level :func:`die_unless_hint_pep586`
    validator to do so before leveraging this tuple elsewhere.

    Parameters
    ----------
    hint : object
        :pep:`586`-compliant type hint to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep586Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    ----------
    tuple
        Tuple of zero or more literal objects subscripting this hint.

    Raises
    ----------
    :exc:`exception_cls`
        If this object is *not* a :pep:`586`-compliant type hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_sign

    # If this hint is *NOT* PEP 586-compliant, raise an exception.
    if get_hint_pep_sign(hint) is not HintSignLiteral:
        raise exception_cls(
            f'{exception_prefix}PEP 586 type hint {repr(hint)} neither '
            f'"typing.Literal" nor "typing_extensions.Literal".'
        )
    # Else, this hint is PEP 586-compliant.

    # Return the standard tuple of all literals subscripting this hint.
    return hint.__args__



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) mungers** (i.e., low-level callables
modifying various properties of various nodes in the currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Attribute,
    Call,
    Index,
    Name,
    Subscript,
    expr,
    keyword,
)
from beartype.claw._clawmagic import (
    NODE_CONTEXT_LOAD,
    BEARTYPE_CLAW_STATE_OBJ_NAME,
    BEARTYPE_CLAW_STATE_CONF_CACHE_VAR_NAME,
    BEARTYPE_DECORATOR_FUNC_NAME,
)
from beartype.claw._clawtyping import (
    NodeDecoratable,
)
from beartype._conf.confcls import (
    BEARTYPE_CONF_DEFAULT,
    BeartypeConf,
)
from beartype._util.ast.utilastmunge import copy_node_metadata
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_9

# ....................{ DECORATORS                         }....................
#FIXME: Unit test us up, please.
def decorate_node(node: NodeDecoratable, conf: BeartypeConf) -> None:
    '''
    Add a new **child beartype decoration node** (i.e., abstract syntax tree
    (AST) node applying the :func:`beartype.beartype` decorator configured by
    the passed beartype configuration) to the passed **parent decoratable node**
    (i.e., AST node encapsulating the definition of a pure-Python object
    supporting decoration by one or more ``"@"``-prefixed decorations, including
    both pure-Python classes *and* callables).

    Note that this function **prepends** rather than appends this child
    decoration node to the beginning of the list of all child decoration nodes
    for this parent decoratable node. Since that list is "stored outermost first
    (i.e. the first in the list will be applied last)", prepending guarantees
    that the beartype decorator will be applied last (i.e., *after* all other
    decorators). This ensures that explicitly configured beartype decorations
    applied to this decoratable by the end user (e.g.,
    ``@beartype(conf=BeartypeConf(...))``) assume precedence over implicitly
    configured beartype decorations applied by this function.

    Parameters
    ----------
    node : AST
        Decoratable node to add a new child beartype decoration node to.
    conf : BeartypeConf
        **Beartype configuration** (i.e., dataclass configuring the
        :mod:`beartype.beartype` decorator for some decoratable object(s)
        decorated by a parent node passing this dataclass to that decorator).
    '''
    assert isinstance(node, AST), f'{repr(node)} not AST node.'
    assert isinstance(conf, BeartypeConf), f'{repr(conf)} not configuration.'

    # Child decoration node decorating that callable by our beartype decorator.
    beartype_decorator: expr = Name(
        id=BEARTYPE_DECORATOR_FUNC_NAME, ctx=NODE_CONTEXT_LOAD)

    # Copy all source code metadata from this parent callable node onto this
    # child decoration node.
    copy_node_metadata(node_src=node, node_trg=beartype_decorator)

    # If the current beartype configuration is *NOT* the default beartype
    # configuration, this configuration is a user-defined beartype configuration
    # which *MUST* be passed to a call to the beartype decorator. Merely
    # referencing this decorator does *NOT* suffice. In this case...
    if conf != BEARTYPE_CONF_DEFAULT:
        # Replace the reference to this decorator defined above with a
        # call to this decorator passed this configuration.
        beartype_decorator = Call(
            func=beartype_decorator,
            args=[],
            # Node encapsulating the passing of this configuration as
            # the "conf" keyword argument to this call.
            keywords=[make_node_keyword_conf(node_sibling=node)],
        )

        # Copy all source code metadata from this parent callable node onto this
        # child call node.
        copy_node_metadata(node_src=node, node_trg=beartype_decorator)
    # Else, this configuration is simply the default beartype configuration. In
    # this case, avoid passing that configuration to the beartype decorator for
    # both efficiency and simplicity.

    # Prepend this child decoration node to the beginning of the list of all
    # child decoration nodes for this parent callable node. Since this list is
    # "stored outermost first (i.e. the first in the list will be applied
    # last)", prepending guarantees that our decorator will be applied last
    # (i.e., *AFTER* all subsequent decorators). This ensures that explicitly
    # configured @beartype decorations (e.g.,
    # "beartype(conf=BeartypeConf(...))") assume precedence over implicitly
    # configured @beartype decorations inserted by this hook.
    node.decorator_list.insert(0, beartype_decorator)

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_node_keyword_conf(node_sibling: AST) -> keyword:
    '''
    Create and return a new **beartype configuration keyword argument node**
    (i.e., abstract syntax tree (AST) node passing the beartype configuration
    associated with the currently visited module as a ``conf`` keyword to a
    :func:`beartype.beartype` decorator orchestrated by the caller).

    Parameters
    ----------
    node_sibling : AST
        Sibling node to copy source code metadata from.

    Returns
    ----------
    keyword
        Keyword node passing this configuration to an arbitrary function.
    '''

    # Node encapsulating a reference to the beartype import hook state global.
    node_claw_state = Name(
        id=BEARTYPE_CLAW_STATE_OBJ_NAME, ctx=NODE_CONTEXT_LOAD)

    # Node encapsulating the fully-qualified name of that module of the
    # currently visited module, accessed via the "__name__" dunder attribute
    # accessible to any module.
    node_module_name = Name(id='__name__', ctx=NODE_CONTEXT_LOAD)

    # Node encapsulating a reference to the beartype configuration object cache
    # (i.e., dictionary mapping from fully-qualified module names to the
    # beartype configurations associated with those modules).
    node_module_name_to_conf = Attribute(
        value=node_claw_state,
        attr=BEARTYPE_CLAW_STATE_CONF_CACHE_VAR_NAME,
        ctx=NODE_CONTEXT_LOAD,
    )

    # Node encapsulating the indexation of a dictionary by the fully-qualified
    # name of that module of the currently visited module.
    node_module_name_index: AST = None  # type: ignore[assignment]

    # If the active Python interpreter targets Python >= 3.9...
    if IS_PYTHON_AT_LEAST_3_9:
        # Reuse this node in a manner specific to Python >= 3.9, which
        # fundamentally broke backward compatibility with Python 3.8 with
        # respect to dictionary subscription.
        node_module_name_index = node_module_name
    # Else, the active Python interpreter targets Python 3.8. In this case..
    else:
        # Create this node in a manner specific to Python 3.8, which requires
        # an additional intermediary node *NOT* required under Python >= 3.9.
        node_module_name_index = Index(value=node_module_name)

        # Copy all source code metadata (e.g., line numbers) from this sibling
        # node onto this new node.
        copy_node_metadata(
            node_src=node_sibling, node_trg=node_module_name_index)

    # Node encapsulating a reference to this beartype configuration, indirectly
    # (and efficiently) accessed via a dictionary lookup into this object cache.
    # While cumbersome, this indirection is effectively "glue" integrating this
    # AST node generation algorithm with the corresponding Python code
    # subsequently interpreted by Python at runtime during module importation.
    node_conf = Subscript(
        value=node_module_name_to_conf,
        slice=node_module_name_index,
        ctx=NODE_CONTEXT_LOAD,
    )

    # Node encapsulating the passing of this beartype configuration by the
    # "conf" keyword argument to an arbitrary function call of some suitable
    # "beartype" function orchestrated by the caller.
    node_keyword_conf = keyword(arg='conf', value=node_conf)

    # Copy all source code metadata (e.g., line numbers) from this sibling node
    # onto these new nodes.
    copy_node_metadata(
        node_src=node_sibling,
        node_trg=(
            node_claw_state,
            node_module_name,
            node_module_name_to_conf,
            node_conf,
            node_keyword_conf,
        )
    )

    # Return this "conf" keyword node.
    return node_keyword_conf



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator type hint code generation unit tests.**

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to type-checking code dynamically generated by the
:mod:`beartype._decor.wrap.wrapmain` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ decorator                  }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Urgh!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
def test_beartype(iter_hints_piths_meta) -> None:
    '''
    Test the :func:`beartype.beartype` decorator with respect to type-checking
    code dynamically generated by the
    :mod:`beartype._decor.wrap.wrapmain` submodule.

    This unit test effectively acts as a functional test and is thus the core
    test exercising decorator functionality from the end user perspective --
    the only perspective that matters in the end. Unsurprisingly, this test is
    mildly more involved than most. *Whatevah.*

    This test additionally attempts to avoid similar issues to a `prior issue
    <issue #5_>`__ of this decorator induced by repeated
    :func:`beartype.beartype` decorations of different callables annotated by
    one or more of the same PEP-compliant type hints.

    .. _issue #5:
       https://github.com/beartype/beartype/issues/5

    Parameters
    ----------
    iter_hints_piths_meta : Callable[[], Iterable[beartype_test.a00_unit.data.hint.util.data_hintmetautil.HintPithMetadata]]
        Factory function creating and returning a generator iteratively yielding
        ``HintPithMetadata`` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating that hint.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintViolation,
        # BeartypeDecorHintPep585DeprecationWarning,
    )
    from beartype._util.hint.pep.utilpeptest import is_hint_pep_deprecated
    from beartype._util.text.utiltextansi import strip_str_ansi
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithUnsatisfiedMetadata,
    )
    from beartype_test._util.pytroar import raises_uncached
    # from pytest import warns
    # from pytest import deprecated_call
    from re import search

    # ....................{ MAIN                           }....................
    # For each predefined type hint and associated metadata...
    for hint_pith_meta in iter_hints_piths_meta():
        # ....................{ LOCALS                     }....................
        # Metadata describing this hint.
        hint_meta = hint_pith_meta.hint_meta

        # Type hint to be type-checked.
        hint = hint_meta.hint

        # Object to type-check against this hint.
        pith = hint_pith_meta.pith

        # Metadata describing this pith.
        pith_meta = hint_pith_meta.pith_meta
        # print(f'Type-checking PEP type hint {repr(hint_meta.hint)}...')

        # Beartype decorator configured specifically for this hint.
        beartype_confed = beartype(conf=hint_meta.conf)

        # Undecorated callable both accepting a single parameter and returning
        # a value annotated by this hint whose implementation trivially reduces
        # to the identity function.
        def func_untyped(hint_param: hint) -> hint:
            return hint_param

        # @beartype-generated wrapper function type-checking this callable.
        func_typed = None

        # If...
        if (
            # This hint is PEP-compliant (rather than PEP-noncompliant) *AND*...
            isinstance(hint_meta, HintPepMetadata) and
            # This hint is deprecated (e.g., by PEP 585)...
            is_hint_pep_deprecated(hint)
        ):
            #FIXME: For unknown and probably uninteresting reasons, the
            #pytest.warns() context manager appears to be broken on our
            #local machine. We have no recourse but to unconditionally
            #ignore this warning at the module level. So much rage!
            #FIXME: It's likely this has something to do with the fact that
            #Python filters deprecation warnings by default. This is almost
            #certainly a pytest issue. Since this has become fairly
            #unctuous, we should probably submit a pytest issue report.
            #FIXME: Actually, pytest now appears to have explicit support for
            #testing that a code block emits a deprecation warning:
            #    with pytest.deprecated_call():
            #        myfunction(17)
            #See also: https://docs.pytest.org/en/6.2.x/warnings.html#ensuring-code-triggers-a-deprecation-warning
            #FIXME: Fascinatingly, warns() still refuses to capture warnings.
            #Although we certainly could call deprecated_call(), doing so is
            #stymied by the fact that
            #"BeartypeDecorHintPep585DeprecationWarning" does *NOT* subclass the
            #standard "DeprecationWarning" class. *sigh*
            func_typed = beartype_confed(func_untyped)

            #FIXME: Pass the "match" parameter to deprecated_call() to assert
            #that @beartype emits the expected deprecation warning.
            # Decorate thes callable under a context manager asserting this
            # declaration to emit non-fatal deprecation warnings.
            # with warns(BeartypeDecorHintPep585DeprecationWarning):
            # with warns():
            # with deprecated_call():
            #     func_typed = beartype_confed(func_untyped)
        # Else, this is *NOT* a deprecated PEP-compliant type hint. In this
        # case, decorated this callable as is.
        else:
            func_typed = beartype_confed(func_untyped)

        # # @beartype-generated wrapper function type-checking this callable.
        # func_typed = (func_untyped)

        # ....................{ VIOLATE                    }....................
        # If this pith violates this hint...
        if isinstance(pith_meta, HintPithUnsatisfiedMetadata):
            # ....................{ EXCEPTION ~ type       }....................
            # Assert this wrapper function raises the expected exception when
            # type-checking this pith against this hint.
            with raises_uncached(BeartypeCallHintViolation) as exception_info:
                func_typed(pith)

            # Exception captured by the prior call to this wrapper function.
            exception = exception_info.value

            # Exception type localized for debuggability. Sadly, the
            # pytest.ExceptionInfo.__repr__() dunder method fails to usefully
            # describe its exception metadata.
            exception_type = exception_info.type

            # Assert this exception metadata describes the expected exception
            # as a sanity check against upstream pytest issues and/or issues
            # with our raises_uncached() context manager.
            assert issubclass(exception_type, BeartypeCallHintViolation)

            # Assert this exception to be public rather than private. The
            # @beartype decorator should *NEVER* raise a private exception.
            assert exception_type.__name__[0] != '_'

            # ....................{ EXCEPTION ~ culprits   }....................
            # Tuple of the culprits responsible for this exception, localized to
            # avoid inefficiently recomputing these culprits on each access.
            exception_culprits = exception.culprits

            # Assert that this tuple both exists *AND* is non-empty.
            assert isinstance(exception_culprits, tuple)
            assert exception_culprits

            # First culprit, which *SHOULD* be either:
            # * If this pith can be weakly referenced, this pith.
            # * Else, this pith *CANNOT* be weakly referenced. In this case,
            #   fallback to the truncated representation of this pith.
            culprit_pith = exception_culprits[0]

            # If the first culprit is a string...
            if isinstance(culprit_pith, str):
                # Assert that this string is non-empty.
                assert culprit_pith
            # Else, the first culprit is *NOT* a string. In this case...
            else:
                # Assert that the first culprit is this pith.
                assert culprit_pith is pith

                # If two or more culprits were responsible...
                if len(exception_culprits) >= 2:
                    # For each culprit following the first...
                    for culprit_nonpith in exception_culprits[1:]:
                        # Assert that this subsequent culprit is *NOT* the pith.
                        assert culprit_nonpith is not pith

            # ....................{ EXCEPTION ~ message    }....................
            # Exception message raised by this wrapper function.
            exception_str = str(exception)
            exception_str = strip_str_ansi(exception_str)
            # print('exception message: {}'.format(exception_str))

            # Assert this exception message is prefixed by a substring
            # identifying the decorated callable with a human-readable label.
            assert exception_str.startswith('Function ')

            # Assert that iterables of uncompiled regular expression expected
            # to match and *NOT* match this message are *NOT* strings, as
            # commonly occurs when accidentally omitting a trailing comma in
            # tuples containing only one string: e.g.,
            # * "('This is a tuple, yo.',)" is a 1-tuple containing one string.
            # * "('This is a string, bro.')" is a string *NOT* contained in a
            #   1-tuple.
            assert not isinstance(
                pith_meta.exception_str_match_regexes, str)
            assert not isinstance(
                pith_meta.exception_str_not_match_regexes, str)

            # For each uncompiled regular expression expected to match this
            # message, assert this expression actually does so.
            #
            # Note that the re.search() rather than re.match() function is
            # called. The latter is rooted at the start of the string and thus
            # *ONLY* matches prefixes, while the former is *NOT* rooted at any
            # string position and thus matches arbitrary substrings as desired.
            for exception_str_match_regex in (
                pith_meta.exception_str_match_regexes):
                assert search(
                    exception_str_match_regex,
                    exception_str,
                ) is not None

            # For each uncompiled regular expression expected to *NOT* match
            # this message, assert this expression actually does so.
            for exception_str_not_match_regex in (
                pith_meta.exception_str_not_match_regexes):
                assert search(
                    exception_str_not_match_regex,
                    exception_str,
                ) is None
        # ....................{ SATISFY                    }....................
        # Else, this pith satisfies this hint. In this case...
        else:
            # Assert this wrapper function successfully type-checks this pith
            # against this hint *WITHOUT* modifying this pith.
            assert func_typed(pith) is pith



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2024 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable getters** (i.e., utility functions dynamically
querying and retrieving various properties of passed callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import (
    Any,
    Callable,
    Optional,
)
from beartype._data.hint.datahinttyping import (
    HintAnnotations,
    TypeException,
)

# ....................{ GETTERS ~ hints                    }....................
#FIXME: Refactor all unsafe access of the low-level "__annotations__" dunder
#attribute to instead call this high-level getter, please.
#FIXME: Unit test us up, please.
def get_func_annotations(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> HintAnnotations:
    '''
    **Annotations** (i.e., dictionary mapping from the name of each annotated
    parameter or return of the passed pure-Python callable to the type hint
    annotating that parameter or return) of that callable.

    Parameters
    ----------
    func : object
        Object to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    ----------
    HintAnnotations
        Annotations of that callable.

    Raises
    ----------
    exception_cls
         If that callable is *not* actually a pure-Python callable.

    See Also
    ----------
    :func:`.get_func_annotations_or_none`
        Further details.
    '''

    # Annotations of that callable if that callable is actually a pure-Python
    # callable *OR* "None" otherwise.
    hint_annotations = get_func_annotations_or_none(func)

    # If that callable is *NOT* pure-Python, raise an exception.
    if hint_annotations is None:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not class.')
        assert issubclass(exception_cls, Exception), (
            f'{repr(exception_cls)} not exception subclass.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # If that callable is uncallable, raise an appropriate exception.
        if not callable(func):
            raise exception_cls(f'{exception_prefix}{repr(func)} not callable.')
        # Else, that callable is callable.

        # Raise a human-readable exception.
        raise exception_cls(
            f'{exception_prefix}{repr(func)} not pure-Python function.')
    # Else, that callable is pure-Python.

    # Return these annotations.
    return hint_annotations


#FIXME: Refactor all unsafe access of the low-level "__annotations__" dunder
#attribute to instead call this high-level getter, please.
#FIXME: Unit test us up, please.
def get_func_annotations_or_none(func: Callable) -> Optional[HintAnnotations]:
    '''
    **Annotations** (i.e., dictionary mapping from the name of each annotated
    parameter or return of the passed pure-Python callable to the type hint
    annotating that parameter or return) of that callable if that callable is
    actually a pure-Python callable *or* :data:`None` otherwise (i.e., if that
    callable is *not* a pure-Python callable).

    Parameters
    ----------
    func : object
        Object to be inspected.

    Returns
    ----------
    Optional[HintAnnotations]
        Either:

        * If that callable is actually a pure-Python callable, the annotations
          of that callable.
        * Else, :data:`None`.
    '''

    # Demonstrable monstrosity demons!
    #
    # Note that the "__annotations__" dunder attribute is guaranteed to exist
    # *ONLY* for standard pure-Python callables. Various other callables of
    # interest (e.g., functions exported by the standard "operator" module) do
    # *NOT* necessarily declare that attribute. Since this tester is commonly
    # called in general-purpose contexts where this guarantee does *NOT*
    # necessarily hold, we intentionally access that attribute safely albeit
    # somewhat more slowly via getattr().
    return getattr(func, '__annotations__', None)

# ....................{ GETTERS ~ descriptor               }....................
def get_func_classmethod_wrappee(
    # Mandatory parameters.
    func: Any,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> Callable:
    '''
    Pure-Python unbound function wrapped by the passed **C-based unbound class
    method descriptor** (i.e., method decorated by the builtin
    :class:`classmethod` decorator, yielding a non-callable instance of that
    :class:`classmethod` decorator class implemented in low-level C and
    accessible via the low-level :attr:`object.__dict__` dictionary rather than
    as class or instance attributes).

    Parameters
    ----------
    func : object
        Object to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised. Defaults to
        :static:`_BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    ----------
    Callable
        Pure-Python unbound function wrapped by this class method descriptor.

    Raises
    ----------
    exception_cls
         If the passed object is *not* a class method descriptor.

    See Also
    ----------
    :func:`beartype._util.func.utilfunctest.is_func_classmethod`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import die_unless_func_classmethod

    # If this object is *NOT* a class method descriptor, raise an exception.
    die_unless_func_classmethod(
        func=func,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this object is a class method descriptor.

    # Return the pure-Python function wrapped by this descriptor. Just do it!
    return func.__func__


def get_func_staticmethod_wrappee(
    # Mandatory parameters.
    func: Any,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> Callable:
    '''
    Pure-Python unbound function wrapped by the passed **C-based unbound static
    method descriptor** (i.e., method decorated by the builtin
    :class:`staticmethod` decorator, yielding a non-callable instance of that
    :class:`staticmethod` decorator class implemented in low-level C and
    accessible via the low-level :attr:`object.__dict__` dictionary rather than
    as class or instance attributes).

    Parameters
    ----------
    func : object
        Object to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised. Defaults to
        :static:`_BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    ----------
    Callable
        Pure-Python unbound function wrapped by this static method descriptor.

    Raises
    ----------
    exception_cls
         If the passed object is *not* a static method descriptor.

    See Also
    ----------
    :func:`beartype._util.func.utilfunctest.is_func_staticmethod`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import die_unless_func_staticmethod

    # If this object is *NOT* a static method descriptor, raise an exception.
    die_unless_func_staticmethod(
        func=func,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this object is a static method descriptor.

    # Return the pure-Python function wrapped by this descriptor. Just do it!
    return func.__func__
